import warnings
from datetime import datetime
from typing import Callable, List, Optional, Union

import bw2data as bd
import matplotlib.pyplot as plt
import pandas as pd


def extract_date_as_integer(dt_obj: datetime, time_res: Optional[str] = "year") -> int:
    """
    Converts a datetime object to an integer for a given temporal resolution `time_res`

    Parameters
    ----------

    dt_obj : Datetime object.
        Datetime object to be converted to an integer.

    time_res : str, optional
        time resolution to be returned: year=YYYY, month=YYYYMM, day=YYYYMMDD, hour=YYYYMMDDHH

    Returns
    -------
    date_as_integer : int
        Datetime objectconverted to an integer in the format of time_res

    """
    time_res_dict = {
        "year": "%Y",
        "month": "%Y%m",
        "day": "%Y%m%d",
        "hour": "%Y%m%d%H",
    }

    if time_res not in time_res_dict.keys():
        warnings.warn(
            'time_res: {} is not a valid option. Please choose from: {} defaulting to "year"'.format(
                time_res, time_res_dict.keys()
            ),
            category=Warning,
        )
    formatted_date = dt_obj.strftime(time_res_dict[time_res])
    date_as_integer = int(formatted_date)

    return date_as_integer


def extract_date_as_string(temporal_grouping: str, timestamp: datetime) -> str:
    """
    Extracts the grouping date as a string from a datetime object, based on the chosen temporal grouping.
    e.g. for `temporal_grouping` = 'month', and `timestamp` = 2023-03-29T01:00:00, it extracts the string '202303'.


    Parameters
    ----------
    temporal_grouping : str
        Temporal grouping for the date string. Options are: 'year', 'month', 'day', 'hour'
    timestamp : datetime
        Datetime object to be converted to a string.

    Returns
    -------
    date_as_string
        Date as a string in the format of the chosen temporal grouping.
    """
    time_res_dict = {
        "year": "%Y",
        "month": "%Y%m",
        "day": "%Y%m%d",
        "hour": "%Y%m%d%H",
    }

    if temporal_grouping not in time_res_dict.keys():
        warnings.warn(
            'temporal_grouping: {} is not a valid option. Please choose from: {} defaulting to "year"'.format(
                temporal_grouping, time_res_dict.keys()
            ),
            category=Warning,
        )
    return timestamp.strftime(time_res_dict[temporal_grouping])


def convert_date_string_to_datetime(temporal_grouping, datestring) -> datetime:
    """
    Converts the string of a date to datetime object.
    e.g. for `temporal_grouping` = 'month', and `datestring` = '202303', it extracts 2023-03-01

    Parameters
    ----------
    temporal_grouping : str
        Temporal grouping for the date string. Options are: 'year', 'month', 'day', 'hour'
    datestring : str
        Date as a string

    Returns
    -------
    datetime
        Datetime object of the date string at the chosen temporal resolution.
    """
    time_res_dict = {
        "year": "%Y",
        "month": "%Y%m",
        "day": "%Y%m%d",
        "hour": "%Y%m%d%H",
    }

    if temporal_grouping not in time_res_dict.keys():
        warnings.warn(
            'temporal grouping: {} is not a valid option. Please choose from: {} defaulting to "year"'.format(
                temporal_grouping, time_res_dict.keys()
            ),
            category=Warning,
        )
    return datetime.strptime(datestring, time_res_dict[temporal_grouping])


def add_flows_to_characterization_function_dict(
    flows: Union[str, List[str]],
    func: Callable,
    characterization_function_dict: Optional[dict] = dict(),
) -> dict:
    """
    Add a new flow or a list of flows to the available characterization functions.

    Parameters
    ----------
    flows : Union[str, List[str]]
        Flow or list of flows to be added to the characterization function dictionary.
    func : Callable
        Dynamic characterization function for flow.
    characterization_function_dict : dict, optional
        Dictionary of flows and their corresponding characterization functions. Default is an empty dictionary.

    Returns
    -------
    dict
        Updated characterization function dictionary with the new flow(s) and function(s).
    """

    # Check if the input is a single flow (str) or a list of flows (List[str])
    if isinstance(flows, str):
        # It's a single flow, add it directly
        characterization_function_dict[flows] = func
    elif isinstance(flows, list):
        # It's a list of flows, iterate and add each one
        for flow in flows:
            characterization_function_dict[flow] = func

    return characterization_function_dict


def plot_characterized_inventory_as_waterfall(
    lca_obj,
    static_scores=None,
    prospective_scores=None,
    order_stacked_activities=None,
):
    """
    Plot a stacked waterfall chart of characterized inventory data. As comparison, static and prospective scores can be added.
    Only works for metric GWP at the moment.

    Parameters
    ----------
    lca_obj : TimexLCA
        LCA object with characterized inventory data.
    static_scores : dict, optional
        Dictionary of static scores. Default is None.
    prospective_scores : dict, optional
        Dictionary of prospective scores. Default is None.
    order_stacked_activities : list, optional
        List of activities to order the stacked bars in the waterfall plot. Default is None.

    Returns
    -------
    None but plots the waterfall chart.

    """
    if not hasattr(lca_obj, "characterized_inventory"):
        raise ValueError("LCA object does not have characterized inventory data.")

    if not hasattr(lca_obj, "activity_time_mapping_dict_reversed"):
        raise ValueError("Make sure to pass an instance of a TimexLCA.")

    # Grouping and summing data
    plot_data = lca_obj.characterized_inventory.groupby(
        ["date", "activity"], as_index=False
    ).sum()
    plot_data["year"] = plot_data[
        "date"
    ].dt.year  # TODO make temporal resolution flexible

    # Optimized activity label fetching
    unique_activities = plot_data["activity"].unique()
    activity_labels = {
        idx: bd.get_activity(lca_obj.activity_time_mapping_dict_reversed[idx][0])[
            "name"
        ]
        for idx in unique_activities
    }
    plot_data["activity_label"] = plot_data["activity"].map(activity_labels)
    # Pivoting data for plotting
    pivoted_data = plot_data.pivot(
        index="year", columns="activity_label", values="amount"
    )

    combined_data = []
    # Adding exchange_scores as a static column
    if static_scores:
        static_data = pd.DataFrame(
            static_scores.items(), columns=["activity_label", "amount"]
        )
        static_data["year"] = "static"
        pivoted_static_data = static_data.pivot(
            index="year", columns="activity_label", values="amount"
        )
        combined_data.append(pivoted_static_data)

    combined_data.append(pivoted_data)  # making sure the order is correct

    # Adding exchange_scores as a prospective column
    if prospective_scores:
        prospective_data = pd.DataFrame(
            prospective_scores.items(), columns=["activity_label", "amount"]
        )
        prospective_data["year"] = "prospective"
        pivoted_prospective_data = prospective_data.pivot(
            index="year", columns="activity_label", values="amount"
        )
        combined_data.append(pivoted_prospective_data)

    combined_df = pd.concat(combined_data, axis=0)

    if order_stacked_activities:
        combined_df = combined_df[
            order_stacked_activities
        ]  # change order of activities in the stacked bars of the waterfall

    # Calculate the bottom for only the dynamic data
    dynamic_bottom = pivoted_data.sum(axis=1).cumsum().shift(1).fillna(0)

    if static_scores and prospective_scores:
        bottom = pd.concat([pd.Series([0]), dynamic_bottom, pd.Series([0])])
    elif static_scores:
        bottom = pd.concat([pd.Series([0]), dynamic_bottom])
    elif prospective_scores:
        bottom = pd.concat([dynamic_bottom, pd.Series([0])])
    else:
        bottom = dynamic_bottom

    # Plotting
    ax = combined_df.plot(
        kind="bar",
        stacked=True,
        bottom=bottom,
        figsize=(14, 6),
        edgecolor="black",
        linewidth=0.5,
    )
    ax.set_ylabel("GWP [kg CO2-eq]")
    ax.set_xlabel("")
    plt.xticks(rotation=45, ha="right")

    if static_scores:
        ax.axvline(x=0.5, color="black", linestyle="--", lw=1)
    if prospective_scores:
        ax.axvline(x=len(combined_df) - 1.5, color="black", linestyle="--", lw=1)

    handles, labels = ax.get_legend_handles_labels()
    ax.legend(handles[::-1], labels[::-1])  # Reversing the order for the legend
    ax.set_axisbelow(True)
    plt.grid(True)
    plt.show()



import warnings
from datetime import datetime
from itertools import chain
from typing import Callable, Optional

import bw2data as bd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sb
from bw2calc import LCA
from bw2data import (
    Database,
    Method,
    Normalization,
    Weighting,
    databases,
    methods,
    normalizations,
    projects,
    weightings,
)
from bw2data.backends.schema import ActivityDataset as AD
from bw2data.backends.schema import get_id
from bw2data.errors import Brightway2Project
from dynamic_characterization import characterize_dynamic_inventory
from peewee import fn
from scipy import sparse

# from .dynamic_biosphere_builder import DynamicBiosphereBuilder
from .dynamic_biosphere_builder import DynamicBiosphereBuilder
from .helper_classes import SetList, TimeMappingDict
from .matrix_modifier import MatrixModifier
from .timeline_builder import TimelineBuilder
from .utils import extract_date_as_integer


class TimexLCA:
    """
    Class to perform time-explicit LCA calculations.

    A TimexLCA retrieves the LCI of processes occuring at explicit points in time and relinks their technosphere
    exchanges to match the technology landscape at that point in time, while keeping track of the timing of the
    resulting emissions. As such, it combines prospective and dynamic LCA approaches.

    TimexLCA first calculates a static LCA, which informs a priority-first graph traversal. From the graph traversal,
    temporal relationships between exchanges and processes are derived. Based on the timing of the processes, bw_timex
    matches the processes at the intersection between foreground and background to the best available background
    databases. This temporal relinking is achieved by using datapackages to add new time-specific processes. The new
    processes and their exchanges to other technosphere processes or biosphere flows extent the technopshere and
    biosphere matrices.

    Temporal information of both processes and biosphere flows are retained, allowing for dynamic LCIA.

    Currently absolute Temporal Distributions for biosphere exchanges are dealt with as a look up function:
    If an activity happens at timestamp X then and the biosphere exchange has an absolute temporal 
    distribution (ATD), it looks up the amount from from the ATD correspnding to timestamp X. 
    E.g.: X = 2024, TD=(data=[2020,2021,2022,2023,2024,.....,2120 ], amount=[3,4,4,5,6,......,3]),
    it will look up the value 6 corresponding 2024. If timestamp X does not exist it find the nearest 
    timestamp available (if two timestamps are equally close, it will take the first in order of
    apearance (see numpy.argmin() for this behabiour). 


    TimexLCA calculates:
     1) a static LCA score (`TimexLCA.static_lca.score`, same as `bw2calc.lca.score`),
     2) a static time-explicit LCA score (`TimexLCA.static_score`), which links LCIs to the respective background databases but without additional temporal dynamics of the biosphere flows,
     3) a dynamic time-explicit LCA score (`TimexLCA.dynamic_score`), with dynamic inventory and dynamic charaterization factors. These are provided for radiative forcing and GWP but can also be user-defined.

    Example
    -------
    >>> demand = {('my_foreground_database', 'my_process'): 1}
    >>> method = ("some_method_family", "some_category", "some_method")
    >>> database_date_dict = {
            'my_background_database_one': datetime.strptime("2020", "%Y"),
            'my_background_database_two': datetime.strptime("2030", "%Y"),
            'my_foreground_database':'dynamic'
        }
    >>> bw_timex = TimexLCA(demand, method, database_date_dict)
    >>> bw_timex.build_timeline() # you can pass many optional arguments here, also for the graph traversal
    >>> bw_timex.lci()
    >>> bw_timex.static_lcia()
    >>> print(bw_timex.static_score)
    >>> bw_timex.dynamic_lcia(metric="radiative_forcing") # different metrics can be used, e.g. "GWP", "radiative_forcing"
    >>> print(bw_timex.dynamic_score)

    """ """"""

    def __init__(
        self,
        demand: dict,
        method: tuple,
        database_date_dict: dict = None,
    ) -> None:
        """
        Instantiating a `TimexLCA` object calculates a static LCA, initializes time mapping dicts for activities and biosphere flows, and stores useful subsets of ids in the node_id_collection_dict.

        Parameters
        ----------
        demand : dict[object: float]
                The demand for which the LCA will be calculated. The keys can be Brightway `Node`
                instances, `(database, code)` tuples, or integer ids.
        method : tuple
                Tuple defining the LCIA method, such as `('foo', 'bar')` or default methods, such as `("EF v3.1", "climate change", "global warming potential (GWP100)")`
        database_date_dict : dict, optional
                Dictionary mapping database names to dates.
        """

        self.demand = demand
        self.method = method
        self.database_date_dict = database_date_dict

        if not self.database_date_dict:
            warnings.warn(
                "No database_date_dict provided. Treating the databases containing the functional unit as dynamic. No remapping to time explicit databases will be done."
            )
            self.database_date_dict = {key[0]: "dynamic" for key in demand.keys()}

        # Create static_only dict that excludes dynamic processes that will be exploded later. This way we only have the "background databases" that we can later link to from the dates of the timeline.
        self.database_date_dict_static_only = {
            k: v for k, v in self.database_date_dict.items() if isinstance(v, datetime)
        }

        # Create some collections of nodes that will be useful down the line, e.g. all nodes from the background databases that link to foregroud nodes.
        self.create_node_id_collection_dict()

        # Calculate static LCA results using a custom prepare_lca_inputs function that includes all background databases in the LCA. We need all the IDs for the time mapping dict.
        fu, data_objs, remapping = self.prepare_static_lca_inputs(
            demand=self.demand, method=self.method
        )
        self.static_lca = LCA(fu, data_objs=data_objs, remapping_dicts=remapping)
        self.static_lca.lci()
        self.static_lca.lcia()

        # Create a time mapping dict that maps each activity to a activity_time_mapping_id in the format (('database', 'code'), datetime_as_integer): time_mapping_id)
        self.activity_time_mapping_dict = TimeMappingDict(
            start_id=bd.backends.ActivityDataset.select(fn.MAX(AD.id)).scalar() + 1
        )  # making sure to use unique ids for the time mapped processes by counting up from the highest current activity id

        # Create a similar dict for the biosphere flows. This is populated by the dynamic_biosphere_builder
        self.biosphere_time_mapping_dict = TimeMappingDict(start_id=0)

    ########################################
    # Main functions to be called by users #
    ########################################

    def build_timeline(
        self,
        temporal_grouping: str = "year",
        interpolation_type: str = "linear",
        edge_filter_function: Callable = None,
        cutoff: float = 1e-9,
        max_calc: float = 1e4,
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Creates a `TimelineBuilder` instance that does the graph traversal (similar to bw_temporalis) and extracts all
        edges with their temporal information. Creates the `TimexLCA.timeline` of exchanges.

        Parameters
        ----------
        temporal_grouping : str, optional
            Time resolution for grouping exchanges over time in the timeline. Default is 'year', other options are
            'month', 'day', 'hour'.
        interpolation_type : str, optional
            Type of interpolation when sourcing the new producers in the time-explicit background databases. Default is
            'linear', which means linear interpolation between the closest 2 databases, other options are 'closest',
            which selects only the closest database.
        edge_filter_function : Callable, optional
            Function to skip edges in the graph traversal. Default is to skip all edges within background databases.
        cutoff: float, optional
            The cutoff value for the graph traversal. Default is 1e-9.
        max_calc: float, optional
            The maximum number of calculations to be performed by the graph traversal. Default is 1e4.
        *args : iterable
            Positional arguments for the graph traversal. for `bw_temporalis.TemporalisLCA` passed to the
            `EdgeExtractor` class, which inherits from `TemporalisLCA`.
            See `bw_temporalis` documentation for more information.
        **kwargs : dict
            Additional keyword arguments for `bw_temporalis.TemporalisLCA` passed to the EdgeExtractor class, which
            inherits from TemporalisLCA.
            See bw_temporalis documentation for more information.

        Returns
        -------
        pandas.DataFrame:
            A DataFrame containing the timeline of exchanges

        See also
        --------
        bw_timex.timeline_builder.TimelineBuilder: Class that builds the timeline.

        """
        if edge_filter_function is None:
            warnings.warn(
                "No edge filter function provided. Skipping all edges within background databases."
            )
            skippable = []
            for db in self.database_date_dict_static_only.keys():
                skippable.extend([node.id for node in bd.Database(db)])
            self.edge_filter_function = lambda x: x in skippable
        else:
            self.edge_filter_function = edge_filter_function

        self.temporal_grouping = temporal_grouping
        self.interpolation_type = interpolation_type
        self.cutoff = cutoff
        self.max_calc = max_calc

        # pre-populate the activity time mapping dict with the static activities.
        # Doing this here because we need the temporal grouping for consistent times resolution.
        self.add_static_activities_to_time_mapping_dict()

        # Create timeline builder that does the graph traversal (similar to bw_temporalis) and extracts
        # all edges with their temporal information. Can later be used to build a timeline with the
        # TimelineBuilder.build_timeline() method.
        self.timeline_builder = TimelineBuilder(
            self.static_lca,
            self.edge_filter_function,
            self.database_date_dict,
            self.database_date_dict_static_only,
            self.activity_time_mapping_dict,
            self.node_id_collection_dict,
            self.temporal_grouping,
            self.interpolation_type,
            self.cutoff,
            self.max_calc,
            *args,
            **kwargs,
        )

        self.timeline = self.timeline_builder.build_timeline()
        return self.timeline[
            [
                "date_producer",
                "producer_name",
                "date_consumer",
                "consumer_name",
                "amount",
                "interpolation_weights",
            ]
        ]

    def lci(
        self,
        build_dynamic_biosphere: Optional[bool] = True,
        expand_technosphere: Optional[bool] = True,
    ) -> None:
        """
        Calculates the time-explicit LCI.

        Generates the biosphere and technosphere modifications via the `MatrixModifier` class by calling `TimexLCA.
        build_datapackage() if `expand_technosphere' is True. Otherwise it generates a dynamic inventory directly from
        from the timeline. Optionally, the dynamic biosphere matrix and dynamic inventory is calculated via
        `TimexLCA.calculate_dynamic_inventory()`. Set `build_dynamic_biosphere` to False if you only want to get a new
        overall score and don't care about the timing
        of the emissions. This saves time and memory.

        Parameters
        ----------
        build_dynamic_biosphere: bool
            if True, build the dynamic biosphere matrix and calculate the dynamic LCI.
            Default is True.
        expand_technosphere: bool
            if True, creates an expanded time-explicit technosphere and calculates the LCI from it. If False, creates no
            new technosphere, but calculates the dynamic inventory directly from the timeline. Building from the timeline
            currently only works if `build_dynamic_biosphere` is also True.

        Returns
        -------
        None, but calls LCI calculations from bw2calc and calculates the dynamic inventory, if `build_dynamic_biosphere` is True.

        See also
        --------
        build_datapackage: Method to create the datapackages that contain the modifications to the technopshere and biosphere matrix using the `MatrixModifier` class.
        calculate_dynamic_inventory: Method to calculate the dynamic inventory if `build_dynamic_biosphere` is True.
        """

        if not expand_technosphere and not build_dynamic_biosphere:
            raise ValueError(
                "Currently it is only possible to calculate a dynamic inventory from the timeline.\
                    Please either set build_dynamic_biosphere=True or expand_technosphere=True"
            )

        if not hasattr(self, "timeline"):
            raise ValueError(
                "Timeline not yet built. Call TimexLCA.build_timeline() first."
            )

        # mapping of the demand id to demand time
        self.demand_timing_dict = self.create_demand_timing_dict()

        self.fu, self.data_objs, self.remapping = self.prepare_bw_timex_inputs(
            demand=self.demand,
            method=self.method,
        )

        if expand_technosphere:
            self.datapackage = self.build_datapackage()
            data_obs = self.data_objs + self.datapackage
            self.expanded_technosphere = True  # set flag for later static lcia useage
        else:  # setup for timeline approach
            self.collect_temporalized_processes_from_timeline()
            data_obs = self.data_objs
            self.expanded_technosphere = False  # set flag for later lcia usage

        self.lca = LCA(
            self.fu,
            data_objs=data_obs,
            remapping_dicts=self.remapping,
        )

        if not build_dynamic_biosphere:
            self.lca.lci()
        else:
            if expand_technosphere:
                self.lca.lci(factorize=True)
                self.calculate_dynamic_inventory(from_timeline=False)
                self.lca.redo_lci(
                    self.fu
                )  # to get back the original LCI - necessary because we do some redo_lci's in the dynamic inventory calculation
            else:
                self.calculate_dynamic_inventory(from_timeline=True)

    def static_lcia(self) -> None:
        """
        Calculates static LCIA using time-explicit LCIs with the standard static characterization factors of the selected LCIA method using `bw2calc.lcia()`.

        Parameters
        ----------
        None

        Returns
        -------
        None, but stores the static score in the attribute `static_score`.
        """
        if not hasattr(self, "lca"):
            raise AttributeError("LCI not yet calculated. Call TimexLCA.lci() first.")
            return
        if not self.expanded_technosphere:
            raise ValueError("Currently the static lcia score can only be calculated if the expanded matrix has been built\
                             Please call TimexLCA.lci(expand_technosphere=True) first.")
        self.lca.lcia()
        self.static_score = self.lca.score

    def dynamic_lcia(
        self,
        metric: str = "radiative_forcing",
        time_horizon: int = 100,
        fixed_time_horizon: bool = False,
        time_horizon_start: datetime = None,
        characterization_function_dict: dict = None,
        characterization_function_co2: dict = None,
    ) -> pd.DataFrame:
        """
        Calculates dynamic LCIA with the `DynamicCharacterization` class using the dynamic inventory and dynamic
        characterization functions. Dynamic characterization is handled by the separate package 
        `dynamic_characterization` (https://dynamic-characterization.readthedocs.io/en/latest/).

        Dynamic characterization functions in the form of a dictionary {biosphere_flow_database_id:
        characterization_function} can be given by the user.
        If none are given, a set of default dynamic characterization functions based on IPCC AR6 are provided from
        `dynamic_characterization` package. These are mapped to the biosphere3 flows of the chosen static climate 
        change impact category. If there is no characterization function for a biosphere flow, it will be ignored.

        Two dynamic climate change metrics are provided: "GWP" and "radiative_forcing".
        The time horizon for the impact assessment can be set with the `time_horizon` parameter, defaulting to 100 years.
        The `fixed_time_horizon` parameter determines whether the emission time horizon for all emissions is calculated from the
        functional unit (`fixed_time_horizon=True`) or from the time of the emission (`fixed_time_horizon=False`).
        The former is the implementation of the Levasseur approach (see https://doi.org/10.1021/es9030003), while the latter is how conventional LCA is done.

        Parameters
        ----------
        metric : str, optional
            the metric for which the dynamic LCIA should be calculated. Default is "radiative_forcing". Available: "GWP" and "radiative_forcing"
        time_horizon: int, optional
            the time horizon for the impact assessment. Unit is years. Default is 100.
        fixed_time_horizon: bool, optional
            Whether the emission time horizon for all emissions is calculated from the functional unit (fixed_time_horizon=True) or from the time of the emission (fixed_time_horizon=False). Default is False.
        time_horizon_start: pd.Timestamp, optional
            The starting timestamp of the time horizon for the dynamic characterization. Only needed for fixed time horizons. Default is datetime.now().
        characterization_function_dict: dict, optional
            Dict of the form {biosphere_flow_database_id: characterization_function}. Default is None, which triggers the use of the provided dynamic characterization functions based on IPCC AR6 Chapter 7.
        characterization_function_co2: Callable, optional
            Characterization function for CO2 emissions. Necessary if GWP metrix is chosen. Default is None, which triggers the use of the provided dynamic characterization function of co2 based on IPCC AR6 Chapter 7.


        Returns
        -------
        pandas.DataFrame
            A Dataframe with the characterized inventory for the chosen metric and parameters. Also stores the sum as attribute `dynamic_score`.

        See also
        --------
        dynamic_charaterization: Package handling the dynamic characterization: https://dynamic-characterization.readthedocs.io/en/latest/
        """

        if not hasattr(self, "dynamic_inventory"):
            raise AttributeError(
                "Dynamic lci not yet calculated. Call TimexLCA.calculate_dynamic_lci() first."
            )

        # Set a default for inventory_in_time_horizon using the full dynamic_inventory_df
        inventory_in_time_horizon = self.dynamic_inventory_df

        # Calculate the latest considered impact date
        t0_date = pd.Timestamp(self.timeline_builder.edge_extractor.t0.date[0])
        latest_considered_impact = t0_date + pd.DateOffset(years=time_horizon)

        # Update inventory_in_time_horizon if a fixed time horizon is used
        if fixed_time_horizon:
            last_emission = self.dynamic_inventory_df.date.max()
            if latest_considered_impact < last_emission:
                warnings.warn(
                    "An emission occurs outside of the specified time horizon and will not be characterized. Please make sure this is intended."
                )
                inventory_in_time_horizon = self.dynamic_inventory_df[
                    self.dynamic_inventory_df.date <= latest_considered_impact
                ]

        if not time_horizon_start:
            time_horizon_start = t0_date

        self.characterized_inventory = characterize_dynamic_inventory(
            dynamic_inventory_df=inventory_in_time_horizon,
            metric=metric,
            characterization_function_dict=characterization_function_dict,
            base_lcia_method=self.method,
            time_horizon=time_horizon,
            fixed_time_horizon=fixed_time_horizon,
            time_horizon_start=time_horizon_start,
            characterization_function_co2=characterization_function_co2,
        )

        self.dynamic_score = self.characterized_inventory["amount"].sum()

        return self.characterized_inventory

    ###############################################
    # Other core functions for the inner workings #
    ###############################################

    def build_datapackage(self) -> list:
        """
        Creates the datapackages that contain the modifications to the technopshere and biosphere matrix using the `MatrixModifier` class.

        Parameters
        ----------
        None

        Returns
        -------
        list
            List of datapackages that contain the modifications to the technopshere and biosphere matrix

        See also
        --------
        bw_timex.matrix_modifier.MatrixModifier: Class that handles the technosphere and biosphere matrix modifications.
        """
        # Create matrix modifier that creates the new datapackages with the exploded processes and new links to background databases.
        self.matrix_modifier = MatrixModifier(
            self.timeline, self.database_date_dict_static_only, self.demand_timing_dict
        )
        self.node_id_collection_dict["temporal_markets"] = (
            self.matrix_modifier.temporal_market_ids
        )
        self.node_id_collection_dict["temporalized_processes"] = (
            self.matrix_modifier.temporalized_process_ids
        )
        return self.matrix_modifier.create_datapackage()

    def calculate_dynamic_inventory(
        self,
        from_timeline: bool = False,
    ) -> None:
        """
        Calculates the dynamic inventory, by first creating a dynamic biosphere matrix using the `DynamicBiosphereBuilder`
        class and then multiplying it with the dynamic supply array. The dynamic inventory matrix is stored in the attribute `dynamic_inventory`.
        It is also converted to a DataFrame and stored in the attribute `dynamic_inventory_df`.

        Parameters
        ----------
        None

        Returns
        -------
        None, but calculates the dynamic inventory and stores it in the attribute `dynamic_inventory` as a matrix and in `dynamic_inventory_df` as a DataFrame.

        See also
        --------
        bw_timex.dynamic_biosphere_builder.DynamicBiosphereBuilder: Class for creating the dynamic biosphere matrix and inventory.
        """

        if not hasattr(self, "lca"):
            warnings.warn(
                "Static Timex LCA has not been run. Call TimexLCA.lci() first."
            )
            return

        self.len_technosphere_dbs = sum(
            [len(bd.Database(db)) for db in self.database_date_dict.keys()]
        )

        self.dynamic_biosphere_builder = DynamicBiosphereBuilder(
            self.lca,
            self.activity_time_mapping_dict,
            self.biosphere_time_mapping_dict,
            self.demand_timing_dict,
            self.node_id_collection_dict,
            self.temporal_grouping,
            self.database_date_dict,
            self.database_date_dict_static_only,
            self.timeline,
            self.interdatabase_activity_mapping,
            from_timeline=from_timeline,
        )
        self.dynamic_biomatrix = (
            self.dynamic_biosphere_builder.build_dynamic_biosphere_matrix(
                from_timeline=from_timeline
            )
        )

        # Build the dynamic inventory
        count = len(self.dynamic_supply_array)
        diagonal_supply_array = sparse.spdiags(
            [self.dynamic_supply_array], [0], count, count
        )  # diagnolization of supply array keeps the dimension of the process, which we want to pass as additional information to the dynamic inventory dict
        self.dynamic_inventory = self.dynamic_biomatrix @ diagonal_supply_array

        self.biosphere_time_mapping_dict_reversed = {
            v: k for k, v in self.biosphere_time_mapping_dict.items()
        }

        self.activity_time_mapping_dict_reversed = {
            v: k for k, v in self.activity_time_mapping_dict.items()
        }
        self.dynamic_inventory_df = self.create_dynamic_inventory_dataframe(from_timeline)

    def create_dynamic_inventory_dataframe(self, from_timeline=False) -> pd.DataFrame:
        """Brings the dynamic inventory from its matrix form in `dynamic_inventory` into the the format
        of a pandas.DataFrame, with the right structure to later apply dynamic characterization functions.

        Format needs to be:

        +------------+--------+------+----------+
        |   date     | amount | flow | activity |
        +============+========+======+==========+
        |  datetime  |   33   |  1   |    2     |
        +------------+--------+------+----------+
        |  datetime  |   32   |  1   |    2     |
        +------------+--------+------+----------+
        |  datetime  |   31   |  1   |    2     |
        +------------+--------+------+----------+

        - date: datetime, e.g. '2024-01-01 00:00:00'
        - flow: flow id
        - activity: activity id

        Parameters
        ----------
        None

        Returns
        -------
        pandas.DataFrame, dynamic inventory in DataFrame format

        """

        dataframe_rows = []
        for i in range(self.dynamic_inventory.shape[0]):
            row_start = self.dynamic_inventory.indptr[i]
            row_end = self.dynamic_inventory.indptr[i + 1]
            for j in range(row_start, row_end):
                row = i
                col = self.dynamic_inventory.indices[j]
                value = self.dynamic_inventory.data[j]
                
                if from_timeline:
                    emitting_process_id = self.timeline.iloc[col]['time_mapped_producer']
                else:
                    emitting_process_id = self.activity_dict.reversed[col]

                bioflow_id, date = self.biosphere_time_mapping_dict_reversed[
                    row
                ]  # indices are already the same as in the matrix, as we create an entirely new biosphere instead of adding new entries (like we do with the technosphere matrix)

                dataframe_rows.append(
                    (
                        date,
                        value,
                        bioflow_id,
                        emitting_process_id,
                    )
                )

        df = pd.DataFrame(
            dataframe_rows, columns=["date", "amount", "flow", "activity"]
        )

        return df.sort_values(by=["date", "amount"], ascending=[True, False])

    #############
    # For setup #
    #############

    def prepare_static_lca_inputs(
        self,
        demand=None,
        method=None,
        weighting=None,
        normalization=None,
        demands=None,
        remapping=True,
        demand_database_last=True,
    ) -> tuple:
        """
        Prepare LCA input arguments in Brightway2.5 style.

        Adapted bw2data.compat.py

        The difference to the original method is that we load all available databases into the matrices instead of just the ones depending on the demand.
        We need this for the creation of the time mapping dict that creates a mapping between the producer id and the reference timing of the databases in the `database_date_dict`.

        Parameters
        ----------
        demand : dict[object: float]
            The demand for which the LCA will be calculated. The keys can be Brightway `Node`
            instances, `(database, code)` tuples, or integer ids.
        method : tuple
            Tuple defining the LCIA method, such as `('foo', 'bar')`. Only needed if not passing
            `data_objs`.
        weighting : tuple
            Tuple defining the LCIA weighting, such as `('foo', 'bar')`. Only needed if not passing
            `data_objs`.
        normalization: str
        demands: list of dicts of demands
        remapping: bool
            If True, remap dictionaries
        demand_database_last: bool
            If True, add the demand databases last in the list `database_names`.

        Returns
        -------
        tuple
            Indexed demand, data objects, and remapping dictionaries

        See also
        --------
        bw2data.compat.prepare_lca_inputs: Original code this function is adapted from (https://github.com/brightway-lca/brightway2-data/blob/main/bw2data/compat.py).

        """
        if not projects.dataset.data.get("25"):
            raise Brightway2Project(
                "Please use `projects.migrate_project_25` before calculating using Brightway 2.5"
            )

        databases.clean()
        data_objs = []
        remapping_dicts = None

        demand_database_names = list(self.database_date_dict.keys())

        if demand_database_names:
            database_names = set.union(
                *[
                    Database(db_label).find_graph_dependents()
                    for db_label in demand_database_names
                ]
            )

            if demand_database_last:
                database_names = [
                    x for x in database_names if x not in demand_database_names
                ] + demand_database_names

            data_objs.extend([Database(obj).datapackage() for obj in database_names])

            if remapping:
                # This is technically wrong - we could have more complicated queries
                # to determine what is truly a product, activity, etc.
                # However, for the default database schema, we know that each node
                # has a unique ID, so this won't produce incorrect responses,
                # just too many values. As the dictionary only exists once, this is
                # not really a problem.
                reversed_mapping = {
                    i: (d, c)
                    for d, c, i in AD.select(AD.database, AD.code, AD.id)
                    .where(AD.database << database_names)
                    .tuples()
                }
                remapping_dicts = {
                    "activity": reversed_mapping,
                    "product": reversed_mapping,
                    "biosphere": reversed_mapping,
                }

        if method:
            assert method in methods
            data_objs.append(Method(method).datapackage())
        if weighting:
            assert weighting in weightings
            data_objs.append(Weighting(weighting).datapackage())
        if normalization:
            assert normalization in normalizations
            data_objs.append(Normalization(normalization).datapackage())

        if demands:
            indexed_demand = [{get_id(k): v for k, v in dct.items()} for dct in demands]
        elif demand:
            indexed_demand = {get_id(k): v for k, v in demand.items()}
        else:
            indexed_demand = None

        return indexed_demand, data_objs, remapping_dicts

    def prepare_bw_timex_inputs(
        self,
        demand=None,
        method=None,
        weighting=None,
        normalization=None,
        demands=None,
        remapping=True,
        demand_database_last=True,
    ) -> tuple:
        """
        Prepare LCA input arguments in Brightway 2.5 style.

        ORIGINALLY FROM bw2data.compat.py

        Changes include:
        - always load all databases in demand_database_names
        - indexed_demand has the id of the new consumer_id of the "exploded" demand

        Parameters
        ----------
        demand : dict[object: float]
            The demand for which the LCA will be calculated. The keys can be Brightway `Node`
            instances, `(database, code)` tuples, or integer ids.
        method : tuple
            Tuple defining the LCIA method, such as `('foo', 'bar')`. Only needed if not passing
            `data_objs`.
        demand_timing_dict: dict
            Dictionary mapping demand ids to their timing.
        weighting : tuple
            Tuple defining the LCIA weighting, such as `('foo', 'bar')`. Only needed if not passing
            `data_objs`.
        normalization: str
        demands: list of dicts of demands
        remapping: bool
            If True, remap dictionaries
        demand_database_last: bool
            If True, add the demand databases last in the list `database_names`.

        Returns
        -------
        tuple
            Indexed demand, data objects, and remapping dictionaries

        See also
        --------
        bw2data.compat.prepare_lca_inputs: Original code this function is adapted from (https://github.com/brightway-lca/brightway2-data/blob/main/bw2data/compat.py).
        """

        if not projects.dataset.data.get("25"):
            raise Brightway2Project(
                "Please use `projects.migrate_project_25` before calculating using Brightway 2.5"
            )

        databases.clean()
        data_objs = []
        remapping_dicts = None

        demand_database_names = list(self.database_date_dict.keys())

        if demand_database_names:
            database_names = set.union(
                *[
                    Database(db_label).find_graph_dependents()
                    for db_label in demand_database_names
                ]
            )

            if demand_database_last:
                database_names = [
                    x for x in database_names if x not in demand_database_names
                ] + demand_database_names

            data_objs.extend([Database(obj).datapackage() for obj in database_names])

            if remapping:
                # This is technically wrong - we could have more complicated queries
                # to determine what is truly a product, activity, etc.
                # However, for the default database schema, we know that each node
                # has a unique ID, so this won't produce incorrect responses,
                # just too many values. As the dictionary only exists once, this is
                # not really a problem.
                reversed_mapping = {
                    i: (d, c)
                    for d, c, i in AD.select(AD.database, AD.code, AD.id)
                    .where(AD.database << database_names)
                    .tuples()
                }
                remapping_dicts = {
                    "activity": reversed_mapping,
                    "product": reversed_mapping,
                    "biosphere": reversed_mapping,
                }

        if method:
            assert method in methods
            data_objs.append(Method(method).datapackage())
        if weighting:
            assert weighting in weightings
            data_objs.append(Weighting(weighting).datapackage())
        if normalization:
            assert normalization in normalizations
            data_objs.append(Normalization(normalization).datapackage())

        if demands:
            indexed_demand = [
                {
                    self.activity_time_mapping_dict[
                        (
                            bd.get_node(id=bd.get_id(k)).key,
                            self.demand_timing_dict[bd.get_id(k)],
                        )
                    ]: v
                    for k, v in dct.items()
                }
                for dct in demands
            ]
        elif demand:
            indexed_demand = {
                self.activity_time_mapping_dict[
                    (
                        bd.get_node(id=bd.get_id(k)).key,
                        self.demand_timing_dict[bd.get_id(k)],
                    )
                ]: v
                for k, v in demand.items()
            }
        else:
            indexed_demand = None

        return indexed_demand, data_objs, remapping_dicts

    def create_node_id_collection_dict(self) -> None:
        """
        Creates a dict of collections of nodes that will be useful down the line, e.g. to determine static nodes for the graph traversal or create the dynamic biosphere matrix.
        Available collections are:

        - ``demand_database_names``: set of database names of the demand processes
        - ``demand_dependent_database_names``: set of database names of all processes that depend on the demand processes
        - ``demand_dependent_background_database_names``: set of database names of all processes that depend on the demand processes and are in the background databases
        - ``demand_dependent_background_node_ids``: set of node ids of all processes that depend on the demand processes and are in the background databases
        - ``foreground_node_ids``: set of node ids of all processes that are not in the background databases
        - ``first_level_background_node_ids_static``: set of node ids of all processes that are in the background databases and are directly linked to the demand processes
        - ``first_level_background_node_ids_all``: like first_level_background_node_ids_static, but includes first level background processes from other time explicit databases.
        - ``first_level_background_node_id_dbs``: dictionary with the first_level_background_node_ids_static as keys returning their database

        It also initiates an instance of SetList which contains all mappings of equivalent activieties across time-specific databases.
        - ``self.interdatabase_activity_mapping``: instance of SetList
        ----------
            None

        Returns
        -------
            None, but adds the `node_id_collection_dict containing` the above-mentioned collections, as well as interdatabase_activity_mapping
        """
        self.node_id_collection_dict = {}
        self.interdatabase_activity_mapping = SetList()

        # Original variable names preserved, set types for performance and uniqueness
        demand_database_names = {
            db
            for db in self.database_date_dict.keys()
            if db not in self.database_date_dict_static_only.keys()
        }
        self.node_id_collection_dict["demand_database_names"] = demand_database_names

        demand_dependent_database_names = set()
        for db in demand_database_names:
            demand_dependent_database_names.update(bd.Database(db).find_dependents())
        self.node_id_collection_dict["demand_dependent_database_names"] = (
            demand_dependent_database_names
        )

        demand_dependent_background_database_names = (
            demand_dependent_database_names & self.database_date_dict_static_only.keys()
        )
        self.node_id_collection_dict["demand_dependent_background_database_names"] = (
            demand_dependent_background_database_names
        )

        demand_dependent_background_node_ids = {
            node.id
            for db in demand_dependent_background_database_names
            for node in bd.Database(db)
        }
        self.node_id_collection_dict["demand_dependent_background_node_ids"] = (
            demand_dependent_background_node_ids
        )

        foreground_node_ids = {
            node.id
            for db in self.database_date_dict.keys()
            if db not in self.database_date_dict_static_only.keys()
            for node in bd.Database(db)
        }
        self.node_id_collection_dict["foreground_node_ids"] = foreground_node_ids

        first_level_background_node_ids_static = set()
        first_level_background_node_ids_all = set()
        for node_id in foreground_node_ids:
            node = bd.get_node(id=node_id)
            for exc in chain(node.technosphere(), node.substitution()):
                if exc.input["database"] in self.database_date_dict_static_only.keys():
                    first_level_background_node_ids_static.add(exc.input.id)
                    act_set = {(exc.input.id, exc.input["database"])}

                    for background_db in self.database_date_dict_static_only.keys():
                        try:
                            other_node = bd.get_node(
                                **{
                                    "database": background_db,
                                    "name": exc.input["name"],
                                    "product": exc.input["reference product"],
                                    "location": exc.input["location"],
                                }
                            )
                            first_level_background_node_ids_all.add(other_node.id)
                            act_set.add((other_node.id, background_db))
                        except KeyError as e:
                            warnings.warn(
                                f"Failed to find process in database {background_db} for name='{exc.input['name']}', reference product='{exc.input['reference product']}', location='{exc.input['location']}': {e}"
                            )
                    self.interdatabase_activity_mapping.add(act_set)

        self.node_id_collection_dict["first_level_background_node_ids_static"] = (
            first_level_background_node_ids_static
        )
        self.node_id_collection_dict["first_level_background_node_ids_all"] = (
            first_level_background_node_ids_all
        )

    def collect_temporalized_processes_from_timeline(self) -> None:
        """
        Prepares the input for the LCA from the timeline.

        Parameters
        ----------
        None

        Returns
        -------
        None, but adds "temporal_markets" and "temporalized_processes" to the
        node_id_colletion_dict based on the timeline.

        """
        unique_producers = (
            self.timeline.groupby(["producer", "time_mapped_producer"])
            .count()
            .index.values
        )

        temporal_market_ids = set()
        temporalized_process_ids = set()

        for producer, time_mapped_producer in unique_producers:
            if (
                bd.get_activity(producer)["database"]
                in self.database_date_dict_static_only.keys()
            ):
                temporal_market_ids.add(time_mapped_producer)
            else:
                temporalized_process_ids.add(time_mapped_producer)

        self.node_id_collection_dict["temporal_markets"] = temporal_market_ids
        self.node_id_collection_dict["temporalized_processes"] = (
            temporalized_process_ids
        )

    def add_static_activities_to_time_mapping_dict(self) -> None:
        """
        Adds all activities from the static LCA to `activity_time_mapping_dict`, an instance of `TimeMappingDict`.
        This gives a unique mapping in the form of (('database', 'code'), datetime_as_integer): time_mapping_id)
        that is later used to uniquely identify time-resolved processes. This is the pre-population of the time
        mapping dict with the static activities. Further time-explicit activities (from other temporalized
        background databases) are added lateron by in the TimelineBuilder.

        Parameters
        ----------
        None

        Returns
        -------
        None but adds the activities to the `activity_time_mapping_dict`
        """
        for idx in self.static_lca.dicts.activity.keys():  # activity ids
            key = self.static_lca.remapping_dicts["activity"][
                idx
            ]  # ('database', 'code')
            time = self.database_date_dict[
                key[0]
            ]  # datetime (or 'dynamic' for foreground processes)
            if isinstance(time, str):  # if 'dynamic', just add the string
                self.activity_time_mapping_dict.add((key, time), unique_id=idx)
            elif isinstance(time, datetime):
                self.activity_time_mapping_dict.add(
                    (key, extract_date_as_integer(time, self.temporal_grouping)),
                    unique_id=idx,
                )  # if datetime, map to the date as integer
            else:
                warnings.warn(f"Time of activity {key} is neither datetime nor str.")

    def create_demand_timing_dict(self) -> dict:
        """
        Generate a dictionary that maps producer (key) to timing (value) for the demands in the product system.
        It searches the timeline for those rows that contain the functional units (demand-processes as producer and -1 as consumer) and returns the time of the demand as an integer.
        Time of demand can have flexible resolution (year=YYYY, month=YYYYMM, day=YYYYMMDD, hour=YYYYMMDDHH) defined in `temporal_grouping`.

        Parameters
        ----------
        None

        Returns
        -------
        Dictionary mapping producer ids to reference timing for the specified demands.
        """
        demand_ids = [bd.get_activity(key).id for key in self.demand.keys()]
        demand_rows = self.timeline[
            self.timeline["producer"].isin(demand_ids)
            & (self.timeline["consumer"] == -1)
        ]
        self.demand_timing_dict = {
            row.producer: row.hash_producer for row in demand_rows.itertuples()
        }
        return self.demand_timing_dict

    ######################################
    # For creating human-friendly output #
    ######################################

    def create_labelled_technosphere_dataframe(self) -> pd.DataFrame:
        """
        Returns the technosphere matrix as a dataframe with comprehensible labels instead of ids.

        Parameters
        ----------
        None

        Returns
        -------
        pd.DataFrame, technosphere matrix as a pandas.DataFrame with comprehensible labels instead of ids.
        """

        df = pd.DataFrame(self.technosphere_matrix.toarray())
        df.rename(  # from matrix id to activity id
            index=self.dicts.activity.reversed,
            columns=self.dicts.activity.reversed,
            inplace=True,
        )
        df.rename(  # from activity id to ((database, code), time)
            index=self.activity_time_mapping_dict.reversed(),
            columns=self.activity_time_mapping_dict.reversed(),
            inplace=True,
        )
        return df

    def create_labelled_biosphere_dataframe(self) -> pd.DataFrame:
        """
        Returns the biosphere matrix as a pandas.DataFrame with comprehensible labels instead of ids.

        Parameters
        ----------
        None

        Returns
        -------
        pd.DataFrame, biosphere matrix as a pandas.DataFrame with comprehensible labels instead of ids.
        """

        df = pd.DataFrame(self.biosphere_matrix.toarray())
        df.rename(  # from matrix id to activity id
            index=self.dicts.biosphere.reversed,
            columns=self.dicts.activity.reversed,
            inplace=True,
        )
        df.rename(
            index=self.remapping_dicts["biosphere"],  # from activity id to bioflow name
            columns=self.activity_time_mapping_dict.reversed(),  # from activity id to ((database, code), time)
            inplace=True,
        )

        return df

    def create_labelled_dynamic_biosphere_dataframe(self) -> pd.DataFrame:
        """
        Returns the dynamic biosphere matrix as a dataframe with comprehensible labels instead of ids.

        Parameters
        ----------
        None

        Returns
        -------
        pd.DataFrame, dynamic biosphere matrix as a pandas.DataFrame with comprehensible labels instead of ids.
        """
        df = pd.DataFrame(self.dynamic_biomatrix.toarray())
        df.rename(  # from matrix id to activity id
            index=self.biosphere_time_mapping_dict_reversed,
            columns=self.dicts.activity.reversed,
            inplace=True,
        )
        df.rename(  # from activity id to ((database, code), time)
            columns=self.activity_time_mapping_dict.reversed(),
            inplace=True,
        )

        df = df.loc[(df != 0).any(axis=1)]  # For readablity, remove all-zero rows

        return df

    def plot_dynamic_inventory(self, bio_flows, cumulative=False) -> None:
        """
        Simple plot of dynamic inventory of a biosphere flow over time, with optional cumulative plotting.

        Parameters
        ----------
        bio_flows : list of int
            database ids of the biosphere flows to plot.
        cumulative : bool
            if True, plot cumulative amounts over time

        Returns
        -------
        none, but shows a plot
        """
        plt.figure(figsize=(14, 6))

        filtered_df = self.dynamic_inventory_df[
            self.dynamic_inventory_df["flow"].isin(bio_flows)
        ]
        aggregated_df = filtered_df.groupby("date").sum()["amount"].reset_index()

        if cumulative:
            aggregated_df["amount"] = np.cumsum(aggregated_df["amount"])

        plt.plot(
            aggregated_df["date"], aggregated_df["amount"], marker="o", linestyle="none"
        )

        plt.ylim(bottom=0)
        plt.xlabel("time")
        plt.ylabel("amount [kg]")
        plt.grid(True)
        plt.tight_layout()  # Adjust layout to make room for the rotated date labels
        plt.show()

    def plot_dynamic_characterized_inventory(
        self,
        cumsum: bool = False,
        sum_emissions_within_activity: bool = False,
        sum_activities: bool = False,
    ) -> None:
        """
        Plot the characterized inventory of the dynamic LCI in a very simple plot.
        Legend and title are selcted automatically based on the chosen metric.

        Parameters
        ----------
        cumsum : bool
            if True, plot cumulative amounts over time
        sum_emissions_within_activity : bool
            if True, sum emissions within each activity over time
        sum_activities : bool
            if True, sum emissions over all activities over time

        Returns
        -------
        none, but shows a plot
        """

        if not hasattr(self, "characterized_inventory"):
            warnings.warn(
                "Characterized inventory not yet calculated. Call TimexLCA.dynamic_lcia() first."
            )
            return

        # Fetch the inventory to use in plotting, modify based on flags
        plot_data = self.characterized_inventory.copy()

        if cumsum:
            plot_data["amount_sum"] = plot_data["amount"].cumsum()
            amount = "amount_sum"
        else:
            amount = "amount"

        if sum_emissions_within_activity:
            plot_data = plot_data.groupby(["date", "activity"]).sum().reset_index()
            plot_data["amount_sum"] = plot_data["amount"].cumsum()

        if sum_activities:
            plot_data = plot_data.groupby("date").sum().reset_index()
            plot_data["amount_sum"] = plot_data["amount"].cumsum()
            plot_data["activity_label"] = "All activities"

        else:  # plotting activities separate

            activity_name_cache = {}

            for activity in plot_data["activity"].unique():
                if activity not in activity_name_cache:
                    activity_name_cache[activity] = bd.get_activity(
                        self.activity_time_mapping_dict_reversed[activity][0]
                    )["name"]

            plot_data["activity_label"] = plot_data["activity"].map(activity_name_cache)

        # Plotting
        plt.figure(figsize=(14, 6))
        axes = sb.scatterplot(x="date", y=amount, hue="activity_label", data=plot_data)

        # Determine y-axis limit flexibly
        if plot_data[amount].min() < 0:
            ymin = plot_data[amount].min() * 1.1
        else:
            ymin = 0

        axes.set_axisbelow(True)
        axes.set_ylim(bottom=ymin)
        axes.set_xlabel("Time")

        handles, labels = axes.get_legend_handles_labels()
        axes.legend(handles[::-1], labels[::-1])
        plt.grid()
        plt.show()

    #########
    # Other #
    #########

    def remap_inventory_dicts(self) -> None:
        """
        Give a warning if users want to use bw25's original mapping function.

        Parameters
        ----------
        None

        Returns
        -------
        None, but gives a warning

        """
        warnings.warn(
            "bw25's original mapping function doesn't work with our new time-mapped matrix entries. The Timex mapping can be found in acvitity_time_mapping_dict and biosphere_time_mapping_dict."
        )

    def __getattr__(self, name):
        """
        Delegate attribute access to the self.lca object if the attribute
        is not found in the TimexLCA instance itself, excluding special attributes.
        """
        if name.startswith("__"):
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{name}'"
            )
        if hasattr(self.lca, name):
            return getattr(self.lca, name)
        if hasattr(self.dynamic_biosphere_builder, name):
            return getattr(self.dynamic_biosphere_builder, name)
        else:
            raise AttributeError(
                f"'TimexLCA' object and its 'lca'- and dynamic_biosphere_builder- attributes have no attribute '{name}'"
            )



"""
Testing if the temporal grouping options (year, month, day and hour) work correctly.
"""

from pathlib import Path
import pytest
import bw2calc as bc
import bw2data as bd
from bw_timex import TimexLCA
from datetime import datetime
from bw_temporalis import TemporalDistribution
import numpy as np


# make sure the test db is loaded
def test_temporal_grouping_db_fixture(temporal_grouping_db):
    assert len(bd.databases) == 4


@pytest.mark.usefixtures("temporal_grouping_db")
class TestClass_Grouping:

    # yearly resolution is alreday tested in test_electric_vehicle.py

    def test_monthly_resolution_score(self):

        self.fu = bd.get_node(database="foreground", code="A")

        database_date_dict = {
            "db_2022": datetime.strptime("2022", "%Y"),
            "db_2024": datetime.strptime("2024", "%Y"),
            "foreground": "dynamic",
        }

        self.tlca = TimexLCA(
            demand={self.fu.key: 1},
            method=("GWP", "example"),
            database_date_dict=database_date_dict,
        )
        self.tlca.build_timeline(
            temporal_grouping="month",
            starting_datetime=datetime.strptime("2024-01-02", "%Y-%m-%d"),
        )
        self.tlca.lci()
        self.tlca.static_lcia()
        print(self.tlca.timeline)

        expected_score = 1 / 3 * 15 + 1 / 6 * 15 + 1 / 6 * 10 + 1 / 3 * 10

        assert self.tlca.score == expected_score

    def test_daily_resolution_score(self):

        # first change TD in database to daily (although not strictly necessary)
        self.fu = bd.get_node(database="foreground", code="A")

        for exc in self.fu.exchanges():
            if exc.input["name"] == "B":
                exc["temporal_distribution"] = TemporalDistribution(
                    np.array(
                        [
                            -15,
                            -10,
                            0,
                        ],
                        dtype="timedelta64[D]",
                    ),
                    np.array([1 / 3, 1 / 3, 1 / 3]),
                )
                exc.save()

        database_date_dict = {
            "db_2022": datetime.strptime(
                "2024-01-10", "%Y-%m-%d"
            ),  # give dbs also a daily resolution (makes calculating the expected score easier)
            "db_2024": datetime.strptime("2024-01-25", "%Y-%m-%d"),
            "foreground": "dynamic",
        }

        self.tlca = TimexLCA(
            demand={self.fu.key: 1},
            method=("GWP", "example"),
            database_date_dict=database_date_dict,
        )
        self.tlca.build_timeline(
            temporal_grouping="day",
            starting_datetime=datetime.strptime("2024-01-25", "%Y-%m-%d"),
        )
        self.tlca.lci()
        self.tlca.static_lcia()
        print(self.tlca.timeline)

        expected_score = 1 / 3 * 15 + 2 / 9 * 15 + 1 / 9 * 10 + 1 / 3 * 10

        assert self.tlca.static_score == pytest.approx(expected_score, rel=0.00005)

    def test_hourly_resolution_score(self):

        # first change TD in database to hourly (although not strictly necessary)
        self.fu = bd.get_node(database="foreground", code="A")

        for exc in self.fu.exchanges():
            if exc.input["name"] == "B":
                exc["temporal_distribution"] = TemporalDistribution(
                    np.array(
                        [
                            -15,
                            -10,
                            0,
                        ],
                        dtype="timedelta64[h]",
                    ),
                    np.array([1 / 3, 1 / 3, 1 / 3]),
                )
                exc.save()

        database_date_dict = {
            "db_2022": datetime.strptime(
                "2024-01-10 06:00", "%Y-%m-%d %H:%M"
            ),  # give dbs also a hourly resolution (makes calculating the expected score easier)
            "db_2024": datetime.strptime("2024-01-10 21:00", "%Y-%m-%d %H:%M"),
            "foreground": "dynamic",
        }

        self.tlca = TimexLCA(
            demand={self.fu.key: 1},
            method=("GWP", "example"),
            database_date_dict=database_date_dict,
        )
        self.tlca.build_timeline(
            temporal_grouping="hour",
            starting_datetime=datetime.strptime("2024-01-10 21:00", "%Y-%m-%d %H:%M"),
        )
        self.tlca.lci()
        self.tlca.static_lcia()
        print(self.tlca.timeline)

        expected_score = 1 / 3 * 15 + 2 / 9 * 15 + 1 / 9 * 10 + 1 / 3 * 10

        assert self.tlca.static_score == pytest.approx(expected_score, rel=0.00005)



"""
Testing the results of the static LCA and the timexLCA for a simple test case of electric vehicle to see if the new interpolated amounts are correct.
"""

from pathlib import Path
import pytest
import bw2calc as bc
import bw2data as bd
from bw_timex import TimexLCA
from datetime import datetime


# make sure the test db is loaded
def test_vehicle_db_fixture(vehicle_db):
    assert len(bd.databases) == 5


# for now, one test class, but could be set up more modularly
@pytest.mark.usefixtures("vehicle_db")
class TestClass_EV:

    @pytest.fixture(autouse=True)
    def setup_method(self, vehicle_db):
        self.electric_vehicle = bd.get_node(database="foreground", code="EV")

        database_date_dict = {
            "db_2020": datetime.strptime("2020", "%Y"),
            "db_2030": datetime.strptime("2030", "%Y"),
            "db_2040": datetime.strptime("2040", "%Y"),
            "foreground": "dynamic",
        }

        self.tlca = TimexLCA(
            demand={self.electric_vehicle.key: 1},
            method=("GWP", "example"),
            database_date_dict=database_date_dict,
        )

        self.tlca.build_timeline(
            starting_datetime=datetime.strptime("2024-01-02", "%Y-%m-%d")
        )
        self.tlca.lci()
        self.tlca.static_lcia()

    def test_static_lca_score(self):
        slca = bc.LCA({self.electric_vehicle.key: 1}, method=("GWP", "example"))
        slca.lci()
        slca.lcia()
        expected_static_score = slca.score

        assert self.tlca.static_lca.score == expected_static_score

    def test_bw_timex_score(self):
        ELECTRICITY_CONSUMPTION = 0.2  # kWh/km
        MILEAGE = 150_000  # km
        LIFETIME = 16  # years

        # Overall mass: 1200 kg
        MASS_GLIDER = 840  # kg
        MASS_POWERTRAIN = 80  # kg
        MASS_BATTERY = 280  # kg

        expected_glider_score = (
            0.6
            * MASS_GLIDER
            * 0.8
            * [
                exc["amount"]
                for exc in bd.get_activity(("db_2020", "glider")).biosphere()
            ][
                0
            ]  # 2022 -> 80% 2020, 20% 2030
            + 0.6
            * MASS_GLIDER
            * 0.2
            * [
                exc["amount"]
                for exc in bd.get_activity(("db_2030", "glider")).biosphere()
            ][0]
            + 0.4
            * MASS_GLIDER
            * 0.7
            * [
                exc["amount"]
                for exc in bd.get_activity(("db_2020", "glider")).biosphere()
            ][
                0
            ]  # 2023 -> 70% 2020, 30% 2030
            + 0.4
            * MASS_GLIDER
            * 0.3
            * [
                exc["amount"]
                for exc in bd.get_activity(("db_2030", "glider")).biosphere()
            ][0]
        )

        expected_powertrain_score = (
            0.6
            * MASS_POWERTRAIN
            * 0.8
            * [
                exc["amount"]
                for exc in bd.get_activity(("db_2020", "powertrain")).biosphere()
            ][
                0
            ]  # 2022 -> 80% 2020, 20% 2030
            + 0.6
            * MASS_POWERTRAIN
            * 0.2
            * [
                exc["amount"]
                for exc in bd.get_activity(("db_2030", "powertrain")).biosphere()
            ][0]
            + 0.4
            * MASS_POWERTRAIN
            * 0.7
            * [
                exc["amount"]
                for exc in bd.get_activity(("db_2020", "powertrain")).biosphere()
            ][
                0
            ]  # 2023 -> 70% 2020, 30% 2030
            + 0.4
            * MASS_POWERTRAIN
            * 0.3
            * [
                exc["amount"]
                for exc in bd.get_activity(("db_2030", "powertrain")).biosphere()
            ][0]
        )

        expected_battery_score = (
            0.6
            * MASS_BATTERY
            * 0.8
            * [
                exc["amount"]
                for exc in bd.get_activity(("db_2020", "battery")).biosphere()
            ][
                0
            ]  # 2022 -> 80% 2020, 20% 2030
            + 0.6
            * MASS_BATTERY
            * 0.2
            * [
                exc["amount"]
                for exc in bd.get_activity(("db_2030", "battery")).biosphere()
            ][0]
            + 0.4
            * MASS_BATTERY
            * 0.7
            * [
                exc["amount"]
                for exc in bd.get_activity(("db_2020", "battery")).biosphere()
            ][
                0
            ]  # 2023 -> 70% 2020, 30% 2030
            + 0.4
            * MASS_BATTERY
            * 0.3
            * [
                exc["amount"]
                for exc in bd.get_activity(("db_2030", "battery")).biosphere()
            ][0]
        )

        expected_electricity_score = (
            MILEAGE
            * ELECTRICITY_CONSUMPTION
            * 0.8
            * [
                exc["amount"]
                for exc in bd.get_activity(("db_2030", "electricity")).biosphere()
            ][
                0
            ]  # electricity in year 2032 -> 80% 2030, 20% 2040
            + MILEAGE
            * ELECTRICITY_CONSUMPTION
            * 0.2
            * [
                exc["amount"]
                for exc in bd.get_activity(("db_2040", "electricity")).biosphere()
            ][0]
        )

        expected_glider_recycling_score = (
            MASS_GLIDER
            * [
                exc["amount"]
                for exc in bd.get_activity(("db_2040", "dismantling")).biosphere()
            ][0]
        )  # dismantling 2041 -> 100% 2040

        expected_battery_recycling_score = -(
            MASS_BATTERY
            * [
                exc["amount"]
                for exc in bd.get_activity(("db_2040", "battery_recycling")).biosphere()
            ][0]
        )  # dismantling 2041 -> 100% 2040, negative sign because of waste flow

        expected_timex_score = (
            expected_glider_score
            + expected_powertrain_score
            + expected_battery_score
            + expected_electricity_score
            + expected_glider_recycling_score
            + expected_battery_recycling_score
        )

        assert self.tlca.static_score == pytest.approx(
            expected_timex_score, abs=0.5
        )  # allow for some rounding errors



import bw2data as bd
import warnings
import pandas as pd
import numpy as np
from typing import Union, Tuple, Optional, Callable
from datetime import datetime, timedelta
from typing import KeysView

from bw_temporalis import TemporalDistribution
from bw2data.configuration import labels
from bw2calc import LCA
from .edge_extractor import EdgeExtractor, Edge
from .utils import (
    extract_date_as_integer,
    extract_date_as_string,
    convert_date_string_to_datetime,
)

class TimelineBuilder:
    """
    This class is responsible for building a timeline of processes based on the temporal relationship of the priority-first graph traversal.

    First, the `EdgeExtractor` is called and it extracts a timeline of exchanges (edge_timeline) with temporal information.
    Identical edges within temporal grouping (e.g. year, month, day, hour) are then grouped together and the amount of exchanges is summed up.
    """

    def __init__(
        self,
        slca: LCA,
        edge_filter_function: Callable,
        database_date_dict: dict,
        database_date_dict_static_only: dict,
        time_mapping_dict: dict,
        node_id_collection_dict: dict,
        temporal_grouping: str = "year",
        interpolation_type: str = "linear",
        cutoff: float = 1e-9,
        max_calc: float = 1e4,
        *args,
        **kwargs,
    ) -> None:
        """
        Parameters
        ----------
        slca: LCA
            A static LCA object.
        edge_filter_function: Callable
            A callable that filters edges. If not provided, a function that always returns False is used.
        database_date_dict: dict
            A dictionary mapping databases to dates.
        time_mapping_dict: dict
          A dictionary to map processes to specific times.
        temporal_grouping: str, optional
            The temporal grouping to be used. Default is "year".
        interpolation_type: str, optional
            The type of interpolation to be used to select the background databases. Default is "linear".
        cutoff: 
            The cutoff value for the graph traversal. Default is 1e-9.
        max_calc: 
            The maximum number of calculations to be performed by the graph traversal. Default is 1e4.
        args:   Variable length argument list
            Keyword arguments passed to the EdgeExtractor which inherits from TemporalisLCA. Here, things like the further settings for graph traversal can be set. For details, see bw_temporalis.TemporalisLCA.
        kwargs: Arbitrary keyword arguments
            Keyword arguments passed to the EdgeExtractor which inherits from TemporalisLCA.
        """
        self.slca = slca
        self.edge_filter_function = edge_filter_function
        self.database_date_dict = database_date_dict
        self.database_date_dict_static_only = database_date_dict_static_only
        self.time_mapping_dict = time_mapping_dict
        self.node_id_collection_dict = node_id_collection_dict
        self.temporal_grouping = temporal_grouping
        self.interpolation_type = interpolation_type
        self.cutoff = cutoff
        self.max_calc = max_calc

        # Finding indices of activities from background databases that are known to be static, i.e. have no temporal distributions connecting to them. These will be be skipped in the graph traversal.
        static_activity_db_indices = [
            node_id
            for node_id in self.node_id_collection_dict[
                "demand_dependent_background_node_ids"
            ]
            if node_id
            not in self.node_id_collection_dict[
                "first_level_background_node_ids_static"
            ]
        ]

        self.edge_extractor = EdgeExtractor(
            slca,
            *args,
            edge_filter_function=edge_filter_function,
            cutoff=self.cutoff,
            max_calc=self.max_calc,
            static_activity_indices=set(static_activity_db_indices),
            **kwargs,
        )
        self.edge_timeline = self.edge_extractor.build_edge_timeline()


    def build_timeline(self) -> pd.DataFrame:
        """
        Create a dataframe with grouped, time-explicit edges and, for each grouped edge, interpolate to the database with the closest time of representativeness.

        Edges from same producer to same consumer that occur at different times within the same time window (temporal_grouping) are grouped together. Possible temporal groupings are "year", "month", "day" and "hour".

        For edges between forground and background system, the column "interpolation weights" assigns the ratio [0-1] of the edge's amount to be taken from the database with the closest time of representativeness. If a process is in the foreground system only, the interpolation weight is set to None.
        Available interpolation types are:
            - "linear": linear interpolation between the two closest databases, based on temporal distance.
            - "closest": closest database is assigned 1

        Parameters
        ----------
        None 
            (all are already passed during instantiation)
        
        Returns
        -------
        pd.DataFrame
            A timeline with grouped, time-explicit edges and interpolation weights to background databases.
        """

        # check if database names match with databases in BW project
        self.check_database_names()

        # Check if temporal_grouping is a valid value
        valid_temporal_groupings = ["year", "month", "day", "hour"]
        if self.temporal_grouping not in valid_temporal_groupings:
            raise ValueError(
                f"Invalid value for 'temporal_grouping'. Allowed values are {valid_temporal_groupings}."
            )

        # Extract edge data into a list of dictionaries
        edges_data = [self.extract_edge_data(edge) for edge in self.edge_timeline]

        # Convert list of dictionaries to dataframe
        edges_df = pd.DataFrame(edges_data)

        # adjust the sign for substitution exchanges:
        edges_df["amount"] = edges_df["amount"] * edges_df["edge_type"].apply(
            lambda x: self.adjust_sign_of_amount_based_on_edge_type(x)
        )

        # Explode datetime and amount columns: each row with multiple dates and amounts is exploded into multiple rows with one date and one amount
        edges_df = edges_df.explode(["consumer_date", "producer_date", "amount"])
        edges_df.drop_duplicates(inplace=True)
        edges_df = edges_df[edges_df["amount"] != 0]

        # For the Functional Unit: set consumer date = producer date as it occurs at the same time
        edges_df.loc[edges_df["consumer"] == -1, "consumer_date"] = edges_df.loc[
            edges_df["consumer"] == -1, "producer_date"
        ]

        # extract grouping time of consumer and producer: processes occuring at different times withing in the same time window of grouping get the same grouping time
        edges_df["consumer_grouping_time"] = edges_df["consumer_date"].apply(
            lambda x: extract_date_as_string(self.temporal_grouping, x)
        )
        edges_df["producer_grouping_time"] = edges_df["producer_date"].apply(
            lambda x: extract_date_as_string(self.temporal_grouping, x)
        )

        # group unique pair of consumer and producer with the same grouping times
        grouped_edges = (
            edges_df.groupby(
                [
                    "producer_grouping_time",
                    "consumer_grouping_time",
                    "producer",
                    "consumer",
                ]
            )
            .agg({"amount": "sum"})
            .reset_index()
        )

        # convert grouping times, which was only used as intermediate variable, back to datetime
        grouped_edges["date_producer"] = grouped_edges["producer_grouping_time"].apply(
            lambda x: convert_date_string_to_datetime(
                self.temporal_grouping, x
            )
        )
        grouped_edges["date_consumer"] = grouped_edges["consumer_grouping_time"].apply(
            lambda x: convert_date_string_to_datetime(
                self.temporal_grouping, x
            )
        )

        # add dates as integers as hashes to the dataframe
        grouped_edges["hash_producer"] = grouped_edges["date_producer"].apply(
            lambda x: extract_date_as_integer(x, time_res=self.temporal_grouping)
        )

        grouped_edges["hash_consumer"] = grouped_edges["date_consumer"].apply(
            lambda x: extract_date_as_integer(x, time_res=self.temporal_grouping)
        )

        # add new processes to time mapping dict
        for row in grouped_edges.itertuples():
            self.time_mapping_dict.add(
                (bd.get_node(id=row.producer).key, row.hash_producer)
            )

        # store the ids from the time_mapping_dict in dataframe
        grouped_edges["time_mapped_producer"] = grouped_edges.apply(
            lambda row: self.time_mapping_dict[
                (bd.get_node(id=row.producer).key, row.hash_producer)
            ],
            axis=1,
        )

        grouped_edges["time_mapped_consumer"] = grouped_edges.apply(
            lambda row: (
                self.time_mapping_dict[
                    (bd.get_node(id=row.consumer).key, row.hash_consumer)
                ]
                if row.consumer != -1
                else -1
            ),
            axis=1,
        )

        # Add interpolation weights to background databases to the dataframe
        grouped_edges = self.add_column_interpolation_weights_to_timeline(
            grouped_edges,
            interpolation_type=self.interpolation_type,
        )

        # Retrieve producer and consumer names
        grouped_edges["producer_name"] = grouped_edges.producer.apply(
            lambda x: bd.get_node(id=x)["name"]
        )
        grouped_edges["consumer_name"] = grouped_edges.consumer.apply(self.get_consumer_name)

        # Reorder columns
        grouped_edges = grouped_edges[
            [
                "hash_producer",
                "time_mapped_producer",
                "date_producer",
                "producer",
                "producer_name",
                "hash_consumer",
                "time_mapped_consumer",
                "date_consumer",
                "consumer",
                "consumer_name",
                "amount",
                "interpolation_weights",
            ]
        ]

        return grouped_edges
    
    ###################################################
    # underlying functions called by build_timeline() #
    ###################################################
       


    def check_database_names(self) -> None:
        """
        Check that the strings of the databases exist in the databases of the brightway project.

        """
        for db in self.database_date_dict_static_only.keys():
            assert (
                db in bd.databases
            ), f"{db} is not in your brightway project databases."
        return
    
    def extract_edge_data(self, edge: Edge) -> dict:
        """
        Stores the attributes of an Edge instance in a dictionary.

        Parameters
        ----------
        edge: Edge 
            Edge instance

        Returns
        -------
        dict
            Dictionary with the attributes of the edge instance.
        """
        try:
            consumer_date = edge.abs_td_consumer.date
            consumer_date = np.array(
                [consumer_date for i in range(len(edge.td_producer))]
            ).T.flatten()
        except AttributeError:
            consumer_date = None
        
        return {
            "producer": edge.producer,
            "consumer": edge.consumer,
            "leaf": edge.leaf,
            "consumer_date": consumer_date,
            "producer_date": edge.abs_td_producer.date,
            "amount": edge.abs_td_producer.amount,
            "edge_type": edge.edge_type,
        }
    
    def adjust_sign_of_amount_based_on_edge_type(self, edge_type):
        """
        It checks if the an exchange is of type substitution or a technosphere exchange, based on bw2data labelling convention, and adjusts the amount accordingly.
        Flips the sign of the amount value in the timeline for substitution (positive technosphere) exchanges.

        Parameters
        ----------
        edge_type: str
            Type of the edge, as defined in the exchange data.

        Returns
        -------
        int
            Multiplier for the amount value, 1 for technosphere exchanges, -1 for substitution exchanges.

        """

        if edge_type in labels.technosphere_negative_edge_types: # variants of technosphere lables
            multiplicator = 1
        elif edge_type in labels.technosphere_positive_edge_types: #variants of production AND substitution labels
            multiplicator = 1
            if edge_type in labels.substitution_edge_types: # overwrite variants of substitution labels
                multiplicator = -1
        else:
            # Raise a warning for unrecognized edge type
            warnings.warn(f"Unrecognized type in this edge: {edge_type}", UserWarning)
            raise ValueError("Unrecognized edge type")
        return multiplicator
            
    def add_column_interpolation_weights_to_timeline(self,
        tl_df: pd.DataFrame,
        interpolation_type: str = "linear",
    ) -> pd.DataFrame:
        """
        Add a column to a timeline with the weights for an interpolation between the two nearest dates, from the list of dates of the available databases.

        Parameters
        ----------
        tl_df: pd.DataFrame 
            Timeline as a dataframe.
        interpolation_type: str, optional
            Type of interpolation between the nearest lower and higher dates. Available options: "linear" and "nearest", defaulting to "linear". 

        Returns
        -------
        pd.DataFrame
            Timeline as a dataframe with a column 'interpolation_weights' added, this column looks like {database_name: weight, database_name: weight}.
        """
        if not self.database_date_dict_static_only:
            tl_df["interpolation_weights"] = None
            warnings.warn(
                "No time-explicit databases are provided. Mapping to time-explicit databases is not possible.",
                category=Warning,
            )
            return tl_df

        dates_list = [
            date
            for date in self.database_date_dict_static_only.values()
            if type(date) == datetime
        ]
        if "date_producer" not in list(tl_df.columns):
            raise ValueError("The timeline does not contain dates.")

        # create reversed dict {date: database} with only static "background" db's
        self.reversed_database_date_dict = {
            v: k
            for k, v in self.database_date_dict_static_only.items()
            if type(v) == datetime
        }

        if self.interpolation_type == "nearest":
            tl_df["interpolation_weights"] = tl_df["date_producer"].apply(
                lambda x: self.find_closest_date(x, dates_list)
            )

        if self.interpolation_type == "linear":
            tl_df["interpolation_weights"] = tl_df["date_producer"].apply(
                lambda x: self.get_weights_for_interpolation_between_nearest_years(
                    x, dates_list, interpolation_type
                )
            )

        else:
            raise ValueError(
                f"Sorry, but {interpolation_type} interpolation is not available yet."
            )

        tl_df["interpolation_weights"] = tl_df.apply(
            self.add_interpolation_weights_at_intersection_to_background, axis=1
        )  # add the weights to the timeline for processes at intersection

        return tl_df
    
    def find_closest_date(self, target: datetime, dates: KeysView[datetime]) -> dict:
        """
        Find the closest date to the target in the dates list.

        Parameters
        ----------
        target : datetime.datetime
            Target datetime object.
        dates : KeysView[datetime]
            List of datetime.datetime objects.

        Returns
        -------
        dict
            Dictionary with the key as the closest datetime.datetime object from the list and a value of 1.
        """

        # If the list is empty, return None
        if not dates:
            return None

        # Sort the dates
        dates = sorted(dates)

        # Use min function with a key based on the absolute difference between the target and each date
        closest = min(dates, key=lambda date: abs(target - date))

        return {closest: 1}
    
    def get_weights_for_interpolation_between_nearest_years(self, 
        reference_date: datetime,
        dates_list: KeysView[datetime],
        interpolation_type: str = "linear",
    ) -> dict:
        """
        Find the nearest dates (lower and higher) for a given date from a list of dates and calculate the interpolation weights based on temporal proximity.

        Parameters
        ----------
        reference_date : datetime
            Target date.
        dates_list : KeysView[datetime]
            List of datetime objects representing the temporal representativeness of the available databases.
        interpolation_type : str, optional
            Type of interpolation between the nearest lower and higher dates. For now, only "linear" is available.

        Returns
        -------
        dict
            Dictionary with datetimes of the available closest databases as keys and the weights for interpolation as values.
        """
        dates_list = sorted(dates_list)

        diff_dates_list = [reference_date - x for x in dates_list]

        if timedelta(0) in diff_dates_list:  # date of process == date of database
            exact_match = dates_list[diff_dates_list.index(timedelta(0))]
            return {exact_match: 1}

        closest_lower = None
        closest_higher = None

        # select the closest lower and higher dates of the database in regards to the date of process
        for date in dates_list:
            if date < reference_date:
                if (
                    closest_lower is None
                    or reference_date - date < reference_date - closest_lower
                ):
                    closest_lower = date
            elif date > reference_date:
                if (
                    closest_higher is None
                    or date - reference_date < closest_higher - reference_date
                ):
                    closest_higher = date

        if closest_lower is None:
            warnings.warn(
                f"Reference date {reference_date} is lower than all provided dates. Data will be taken from the closest higher year.",
                category=Warning,
            )
            return {closest_higher: 1}

        if closest_higher is None:
            warnings.warn(
                f"Reference date {reference_date} is higher than all provided dates. Data will be taken from the closest lower year.",
                category=Warning,
            )
            return {closest_lower: 1}

        if self.interpolation_type == "linear":
            weight = int((reference_date - closest_lower).total_seconds()) / int(
                (closest_higher - closest_lower).total_seconds()
            )
        else:
            raise ValueError(
                f"Sorry, but {interpolation_type} interpolation is not available yet."
            )
        return {closest_lower: 1 - weight, closest_higher: weight}
    
    def add_interpolation_weights_at_intersection_to_background(self, row) -> Union[dict, None]:
        """
        returns the interpolation weights to background databases only for those exchanges, where the producing process
        actually comes from a background database (temporal markets).

        Only these processes are receiving inputs from the background databases.
        All other process in the timeline are not directly linked to the background, so the interpolation weight info is not needed and set to None

        Parameters
        ----------
        row : pd.Series
            Row of the timeline dataframe
        Returns
        -------
        dict
            Dictionary with the name of databases and interpolation weights.
        """

        if (
            row["producer"]
            in self.node_id_collection_dict[
                "first_level_background_node_ids_static"
            ]
        ):
            return {
                self.reversed_database_date_dict[x]: v
                for x, v in row["interpolation_weights"].items()
            }
        else:
            return None
   
    def get_consumer_name(self, id: int) -> str:
        """
        Returns the name of consumer node.
        If consuming node is the functional unit, returns -1.

        Parameters
        ----------
        id : int
            Id of node.

        Returns
        -------
        str
            Name of the node or -1
        """
        try:
            return bd.get_node(id=id)["name"]
        except:
            return "-1"  # functional unit


from datetime import datetime
import numpy as np
import pandas as pd
import bw2data as bd
import pytest

from bw_timex import TimexLCA

# make sure the test db is loaded
def test_dynamic_biomatrix_fixture(dynamic_biomatrix_db):
    assert len(bd.databases) == 3


@pytest.mark.usefixtures("dynamic_biomatrix_db")
class TestClass_dynamic_biomatrix:

    @pytest.fixture(autouse=True)
    def setup_method(self):
        self.node_a = bd.get_node(database="foreground", code="A")

        database_date_dict = { 
            "db_2020": datetime.strptime("2020", "%Y"),
            "foreground": "dynamic",
        }

        self.tlca = TimexLCA(
            demand={self.node_a: 1},
            method=("GWP", "example"),
            database_date_dict=database_date_dict,
        )

        self.tlca.build_timeline(starting_datetime=datetime.strptime("2024-01-02", "%Y-%m-%d"))
        self.tlca.lci(expand_technosphere=True, build_dynamic_biosphere=True)
        self.tlca.static_lcia()

    def test_identical_time_mapped_producers_exist_in_timeline(self):
        
        duplicates = self.tlca.timeline["time_mapped_producer"].duplicated().any()
        assert duplicates, f"No duplicates found in column {"time_mapped_producer"}"
      
    def test_dynamic_biomatrix_for_multiple_identical_time_mapped_producers_in_timeline(self):

        expected_dynamic_biomatrix_as_array = np.array([[0. , 0. , 0. , 0. , 0. , 0. , 0. , 1.5]]) #1.5 CO2 is only added once for the time_mapped producer, depsite it occuring twice in the timeline! Before #78, it was added twice , resulting in 3.0                        

        assert np.array_equal(self.tlca.dynamic_biomatrix.toarray(), expected_dynamic_biomatrix_as_array),  "Arrays are not equal"




import bw2data as bd
import numpy as np
import pandas as pd
from bw2calc import LCA
from bw_temporalis import TemporalDistribution
from scipy import sparse as sp

from .helper_classes import SetList
from .utils import convert_date_string_to_datetime


class DynamicBiosphereBuilder:
    """
    This class is used to build a dynamic biosphere matrix, which in contrast to the normal biosphere matrix has rows for each biosphere flow at their time of emission.
    Thus, the dimensions are (bio_flows at a specific timestep) x (processes).
    """

    def __init__(
        self,
        lca_obj: LCA,
        activity_time_mapping_dict: dict,
        biosphere_time_mapping_dict: dict,
        demand_timing_dict: dict,
        node_id_collection_dict: dict,
        temporal_grouping: str,
        database_date_dict: dict,
        database_date_dict_static_only: dict,
        timeline: pd.DataFrame,
        interdatabase_activity_mapping: SetList,
        from_timeline: bool = False,
    ) -> None:
        """
        Initializes the DynamicBiosphereBuilder object.

        Parameters
        ----------
        lca_obj : LCA objecct
            instance of the bw2calc LCA class, e.g. TimexLCA.lca
        activity_time_mapping_dict : dict
            A dictionary mapping activity to their respective timing in the format (('database', 'code'), datetime_as_integer): time_mapping_id)
        biosphere_time_mapping_dict : dict
            A dictionary mapping biosphere flows to their respective timing in the format (('database', 'code'), datetime_as_integer): time_mapping_id), at this point still empty.
        demand_timing_dict : dict
            A dictionary mapping of the demand to demand time
        node_id_collection_dict : dict
            A dictionary containing lists of node ids for different node subsets
        temporal_grouping : str
            A string indicating the temporal grouping of the processes, e.g. 'year', 'month', 'day', 'hour'
        database_date_dict : dict
            A dictionary mapping database names to their respective date
        database_date_dict_static_only : dict
            A dictionary mapping database names to their respective date, but only containing static databases, which are the background databases.
        timeline: pd.DataFrame
            The edge timeline, created from TimexLCA.build_timeline()
        interdatabase_activity_mapping : SetList
            A list of sets, where each set contains the activity ids of the same activity in different databases
        from_timeline : bool, optional
            A boolean indicating if the dynamic biosphere matrix is built directly from the timeline. Default is False.

        Returns
        -------
        None

        """

        self._time_res_dict = {
            "year": "datetime64[Y]",
            "month": "datetime64[M]",
            "day": "datetime64[D]",
            "hour": "datetime64[h]",
        }

        self.lca_obj = lca_obj
        if not from_timeline:
            self.technosphere_matrix = (
                lca_obj.technosphere_matrix.tocsc()
            )  # convert to csc as this is only used for column slicing
            self.dynamic_supply_array = lca_obj.supply_array
            self.activity_dict = lca_obj.dicts.activity
        else:
            self.dynamic_supply_array = timeline.amount.values.astype(float)
        self.activity_time_mapping_dict = activity_time_mapping_dict
        self.biosphere_time_mapping_dict = biosphere_time_mapping_dict
        self.demand_timing_dict = demand_timing_dict
        self.node_id_collection_dict = node_id_collection_dict
        self.time_res = self._time_res_dict[temporal_grouping]
        self.temporal_grouping = temporal_grouping
        self.database_date_dict = database_date_dict
        self.database_date_dict_static_only = database_date_dict_static_only
        self.timeline = timeline
        self.interdatabase_activity_mapping = interdatabase_activity_mapping
        self.rows = []
        self.cols = []
        self.values = []
        self.unique_rows_cols = set()   # To keep track of (row, col) pairs

    def build_dynamic_biosphere_matrix(
        self,
        from_timeline: bool = False,
    ):
        """
        This function creates a separate biosphere matrix, with the dimenions (bio_flows at a specific timestep) x (processes).
        Every temporally resolved biosphere flow has its own row in the matrix, making it highly sparse.
        The timing of the emitting process and potential additional temporal information of the bioshpere flow (e.g. delay of emission compared to timing of process) are considered.

        Absolute Temporal Distributions for biosphere exchanges are dealt with as a look up function:
        If an activity happens at timestamp X then and the biosphere exchange has an absolute temporal 
        distribution (ATD), it looks up the amount from from the ATD correspnding to timestamp X. 
        E.g.: X = 2024, TD=(data=[2020,2021,2022,2023,2024,.....,2120 ], amount=[3,4,4,5,6,......,3]),
        it will look up the value 6 corresponding 2024. If timestamp X does not exist it find the nearest 
        timestamp available (if two timestamps are equally close, it will take the first in order of
        apearance (see numpy.argmin() for this behabiour). 


        Parameters
        ----------
        from_timeline

        Returns
        -------
        dynamic_biomatrix : scipy.sparse.csr_matrix
            A sparse matrix with the dimensions (bio_flows at a specific timestep) x (processes), where every row represents a biosphere flow at a specific time.
        """

        for row in self.timeline.itertuples():
            idx = row.time_mapped_producer
            if from_timeline:
                process_col_index = row.Index
            else:
                process_col_index = self.activity_dict[
                    idx
                ]  # get the matrix column index

            (
                (original_db, original_code),
                time,
            ) = self.activity_time_mapping_dict.reversed()[  # time is here an integer, with various length depending on temporal grouping, e.g. [Y] -> 2024, [M] - > 202401
                idx
            ]
            if idx in self.node_id_collection_dict["temporalized_processes"]:

                time_in_datetime = convert_date_string_to_datetime(
                    self.temporal_grouping, str(time)
                )  # now time is a datetime

                td_producer = TemporalDistribution(
                    date=np.array([time_in_datetime], dtype=self.time_res),
                    amount=np.array([1]),
                ).date
                date = td_producer[0]

                act = bd.get_node(database=original_db, code=original_code)

                for exc in act.biosphere():
                    if exc.get("temporal_distribution"):
                        td_dates = exc["temporal_distribution"].date 
                        td_values = exc["temporal_distribution"].amount
                        if type(td_dates[0])==np.datetime64:  # If the biosphere flows have an absolute TD, this means we have to look up the biosphere flow for the activity time (td_producer)
                            dates = td_producer  # datetime array, same time as producer
                            values = [
                                exc["amount"] * td_values[
                                    np.argmin(
                                        np.abs(
                                            td_dates.astype(self.time_res)-td_producer.astype(self.time_res)
                                        )
                                    )
                                ]
                              ]  # look up the value correponding to the absolute producer time
                        else:
                            dates = (
                                td_producer + td_dates
                            )  # we can add a datetime of length 1 to a timedelta of length N without problems
                            values = exc["amount"] * td_values

                    else:  # exchange has no TD
                        dates = td_producer  # datetime array, same time as producer
                        values = [exc["amount"]]

                    # Add entries to dynamic bio matrix
                    for date, amount in zip(dates, values):

                        # first create a row index for the tuple((db, bioflow), date))
                        time_mapped_matrix_id = self.biosphere_time_mapping_dict.add(
                            (exc.input.id, date)
                        )

                        # populate lists with which sparse matrix is constructed
                        self.add_matrix_entry_for_biosphere_flows(
                            row=time_mapped_matrix_id,
                            col=process_col_index,
                            amount=amount,
                        )
            elif idx in self.node_id_collection_dict["temporal_markets"]:
                if from_timeline:
                    demand = self.demand_from_timeline(row, original_db)
                else:
                    demand = self.demand_from_technosphere(idx, process_col_index)

                self.lca_obj.redo_lci(demand)

                aggregated_inventory = self.lca_obj.inventory.sum(
                    axis=1
                )  # aggregated biosphere flows of background supply chain emissions. Rows are bioflows.

                for row_idx, amount in enumerate(aggregated_inventory.A1):
                    bioflow = self.lca_obj.dicts.biosphere.reversed[row_idx]
                    ((_, _), time) = self.activity_time_mapping_dict.reversed()[idx]

                    time_in_datetime = convert_date_string_to_datetime(
                        self.temporal_grouping, str(time)
                    )  # now time is a datetime

                    td_producer = TemporalDistribution(
                        date=np.array([str(time_in_datetime)], dtype=self.time_res),
                        amount=np.array([1]),
                    ).date  # TODO: Simplify
                    date = td_producer[0]

                    time_mapped_matrix_id = self.biosphere_time_mapping_dict.add(
                        (bioflow, date)
                    )

                    self.add_matrix_entry_for_biosphere_flows(
                        row=time_mapped_matrix_id, col=process_col_index, amount=amount
                    )

        # now build the dynamic biosphere matrix
        if from_timeline:
            ncols = len(self.timeline)
        else:
            ncols = len(self.activity_time_mapping_dict)
        shape = (max(self.rows) + 1, ncols)
        dynamic_biomatrix = sp.coo_matrix((self.values, (self.rows, self.cols)), shape)
        self.dynamic_biomatrix = dynamic_biomatrix.tocsr()

        return self.dynamic_biomatrix

    def demand_from_timeline(self, row, original_db):
        """
        Returns a demand dict directly from the timeline row
        and its interpolation_weights.
        """
        demand = {}
        for db, amount in row.interpolation_weights.items():
            # if not db in act_time_combinations.get(original_code):  #check if act time combination already exists
            [(timed_act_id, _)] = [
                (act, db_name)
                for (act, db_name) in self.interdatabase_activity_mapping[
                    (row.producer, original_db)
                ]
                if db == db_name
            ]
            # t_act = bd.get_activity(timed_act_id)
            demand[timed_act_id] = amount
        return demand

    def demand_from_technosphere(self, idx, process_col_index):
        """
        Returns a demand dict based on the technosphere colummn.
        """
        technosphere_column = (
            self.technosphere_matrix[:, process_col_index].toarray().ravel()
        )  # 1-d np.array
        demand = {}
        for row_idx, amount in enumerate(technosphere_column):
            if row_idx == self.activity_dict[idx]:  # Skip production exchange
                continue
            if amount == 0:
                continue

            node_id = self.activity_dict.reversed[row_idx]

            if (
                node_id in self.node_id_collection_dict["foreground_node_ids"]
            ):  # We only aggregate background process bioflows
                continue

            # demand[bd.get_node(id=node_id)] = -amount
            demand[node_id] = -amount
        return demand

    def add_matrix_entry_for_biosphere_flows(self, row, col, amount):
        """
        Adds an entry to the lists of row, col and values, which are then used to construct the dynamic biosphere matrix.
        Only unqiue entries are added, i.e. if the same row and col index already exists, the value is not added again. 

        Parameters
        ----------
        row : int
            A row index of a new element to the dynamic biosphere matrix
        col: int
            A column index of a new element to the dynamic biosphere matrix
        amount: float
            The amount of the new element to the dynamic biosphere matrix

        Returns
        -------
        None, but the lists of row, col and values are updated

        """

        if (row, col) not in self.unique_rows_cols:
            self.rows.append(row)
            self.cols.append(col)
            self.values.append(amount)

            self.unique_rows_cols.add((row, col))





import warnings

class SetList:
    """A little helper class fot the mapping of the same/mapped activity in different (temporal) databases.
    It contains a list of sets, that hold can hold the set of tuples of (act_id, database). It is built by
    adding sets to the list, and returns the matching sets if one calls the an item from a set. 

    Example: If the class instance is called my_setlist, my_setlist.add(set). When called """

    def __init__(
        self,
    ) -> None:
        self.list = []

    def add(
        self,
        new_set: set,
    ) -> None:
        """
        This method adds a set to the SetList instance.
        Input
        -----
        new_set: a set to add to the SetList instance
        
        Returns
        -------
        None (or rather nothing)
        
        """

        if new_set not in self.list:
            checklist_items = [item for itemset in self.list for item in new_set if item in itemset]
            checklist_sets = [itemset for itemset in self.list for item in new_set if item in itemset]
            if len(checklist_items) != 0:
                warnings.warn('tried to add {} to the SetList\n, but {} already exist in the SetList in:\n {}. \n Skipping {}'.format(
                    new_set, checklist_items, checklist_sets, new_set))
                pass
            else:
                self.list.append(new_set)
        else:
            pass
    
    def __getitem__(
        self,
        key: any
    ) -> set:
        """
        Returns all sets in the SetList instance containing the key
        Inputs
        ------
        key: the key to look for in the sets of the SetList
        
        Returns
        -------
        A list containing the set or all sets 
        """
        sets = [matching_set for matching_set in self.list if key in matching_set]
        if len(sets)>1:
            warnings.warn('Key found in multiple sets!!!! {} Please check! Returning only the first set'.format(sets))
        if len(sets)>0: 
            return sets[0]
        else: 
            warnings.warn('Key {} not found in SetList'.format(key))
            return None
    
    def __len__(
        self,
    ) -> int:
        return len(self.list)

    def __repr__(self):
        return f"SetList({self.list})"


class TimeMappingDict(dict):
    """
    This class is used to create a dictionary that maps a tuple of (flow and timestamp) to an unique integer id.
    """

    def __init__(self, start_id=2, *args, **kwargs) -> None:
        """"
        Initializes the `TimeMappingDict` object.

        Parameters
        ----------
        start_id : int, optional
            The starting id for the mapping. Default is 2.
        *args : Variable length argument list
        **kwargs : Arbitrary keyword arguments

        Returns
        -------
        None
        """

        super().__init__(*args, **kwargs)
        self._current_id = start_id
        self._check_id = (
            start_id - 1
        )  # check_id that is different from the start id for the reversed dict

    def add(self, process_time_tuple, unique_id=None):
        """"
        Adds a new process_time_tuple to the `TimeMappingDict` object.

        Parameters
        ----------
        process_time_tuple : tuple
            A tuple of (flow and timestamp)
        unique_id : int, optional
            An unique id for the process_time_tuple. Default is None.

        Returns
        -------
        int
            An unique id for the process_time_tuple, and adds it to the dictionary, if not already present.
        """
        if process_time_tuple in self:
            return self[process_time_tuple]

        if unique_id is not None:
            self[process_time_tuple] = unique_id
            return unique_id

        self[process_time_tuple] = self._current_id
        self._current_id += 1

        return self._current_id - 1

    def reversed(self):
        """return a reversed version of dict, update if necessary

        Parameters
        ----------
        None

        Returns
        -------
        dict
            A reversed dictionary of the `TimeMappingDict` object.
        """


        if self._check_id != self._current_id:
            self.reversed_dict = {v: k for k, v in self.items()}
            self._check_id = self._current_id
        return self.reversed_dict



import uuid
from typing import Optional

import bw2data as bd
import bw_processing as bwp
import numpy as np
import pandas as pd


class MatrixModifier:
    """
    This class is responsible for modifying the original LCA matrices to contain the time-explicit processes and to relink them to the time-explicit background databasess.pip

    It does this by creating datapackages that add or change matrix entries in technosphere and biosphere matrices, based on a timeline dataframe (created from TimelineBuilder.build_timeline()).

    """

    def __init__(
        self,
        timeline: pd.DataFrame,
        database_date_dict_static_only: dict,
        demand_timing: dict,
        name: Optional[str] = None,
    ) -> None:
        """
        Initializes the MatrixModifier object and creates empty sets to collect the ids of temporalized processes and temporal markets.

        Parameters
        ----------
        timeline : pd.DataFrame
            A DataFrame of the timeline of exchanges
        database_date_dict_static_only : dict
            A dictionary mapping the static background databases to dates.
        demand_timing : dict
            A dictionary mapping the demand to its timing.
        name : str, optional
            An optional name for the MatrixModifier instance. Default is None.

        """

        self.timeline = timeline
        self.database_date_dict_static_only = database_date_dict_static_only
        self.demand_timing = demand_timing
        self.name = name
        self.temporalized_process_ids = set()
        self.temporal_market_ids = set()

    def create_datapackage(self) -> None:
        """
        Creates a list of datapackages for the technosphere and biosphere matrices, by calling the respective functions.

        Parameters
        ----------
        None

        Returns
        -------
        list
            A list of the technosphere and biosphere datapackages.
        """
        technosphere_datapackage = self.create_technosphere_datapackage()
        biosphere_datapackage = self.create_biosphere_datapackage()

        return [technosphere_datapackage, biosphere_datapackage]

    def create_technosphere_datapackage(self) -> bwp.Datapackage:
        """
        Creates the modifications to the technosphere matrix in form of a datapackage. Datapackages add or overwrite datapoints in the LCA matrices before LCA calculations.
        The technospher datapackage add the temporalized processes from the timeline to the technosphere matrix.

        The heavy lifting of this method happens in the method `add_row_to_datapackage()` that is called.
        Here, each node with a temporal distribution is "exploded", which means each occurrence of this node (e.g. steel production on 2020-01-01
        and steel production on 2015-01-01) becomes a separate, time-explicit new node, by adding the new elements to the technosphere matrix.
        For processes at the intersection with background databases, the timing of the exchanges determines which background database to link to in so called "Temporal Markets".

        Parameters
        ----------
        None

        Returns
        -------
        bwp.Datapackage
            A datapackage containing the modifications for the technosphere matrix.
        """
        datapackage = bwp.create_datapackage(
            sum_inter_duplicates=False
        )  # 'sum_inter_duplicates=False': If the same market is used by multiple foreground processes, the market gets created again, inputs should not be summed.

        new_nodes = set()

        for row in self.timeline.iloc[::-1].itertuples():
            self.add_row_to_datapackage(
                row,
                datapackage,
                new_nodes,
            )

        # Adding the production exchanges for new nodes
        for node_id, production_amount in new_nodes:
            datapackage.add_persistent_vector(
                matrix="technosphere_matrix",
                name=uuid.uuid4().hex,
                data_array=np.array([production_amount], dtype=float),
                indices_array=np.array([(node_id, node_id)], dtype=bwp.INDICES_DTYPE),
            )

        return datapackage

    def create_biosphere_datapackage(self) -> bwp.Datapackage:
        """
        Creates the modifications to the biosphere matrix in form of a datapackage. Datapackages add or overwrite datapoints in the LCA matrices before LCA calculations.
        It adds the biosphere flows to the exploded technosphere processes.

        This function iterates over each unique producer, and for each biosphere exchange of the original activity,
        it creates a new biosphere exchange for the new "temporalized" node.

        Temporal markets have no biosphere exchanges, as they only divide the amount of a technosphere exchange between the different databases.

        Parameters
        ----------
        None

        Returns
        -------
        bwp.Datapackage
            A datapackage containing the modifications for the biosphere matrix.
        """
        unique_producers = (
            self.timeline.groupby(["producer", "time_mapped_producer"])
            .count()
            .index.values
        )  # array of unique (producer, timestamp) tuples

        datapackage_bio = bwp.create_datapackage(sum_inter_duplicates=False)

        for producer in unique_producers:
            if (
                bd.get_activity(producer[0])["database"]
                not in self.database_date_dict_static_only.keys()  # skip temporal markets
            ):
                producer_id = producer[1]
                # the producer_id is a combination of the activity_id and the timestamp
                producer_node = bd.get_node(id=producer[0])
                indices = (
                    []
                )  # list of (biosphere, technosphere) indices for the biosphere flow exchanges
                amounts = []  # list of amounts corresponding to the bioflows
                for exc in producer_node.biosphere():
                    indices.append(
                        (exc.input.id, producer_id)
                    )  # directly build a list of tuples to pass into the datapackage, the producer_id is used to for the column of that activity
                    amounts.append(exc.amount)

                datapackage_bio.add_persistent_vector(
                    matrix="biosphere_matrix",
                    name=uuid.uuid4().hex,
                    data_array=np.array(amounts, dtype=float),
                    indices_array=np.array(
                        indices,
                        dtype=bwp.INDICES_DTYPE,
                    ),
                    flip_array=np.array([False], dtype=bool),
                )
        return datapackage_bio

    def add_row_to_datapackage(
        self,
        row: pd.core.frame,
        datapackage: bwp.Datapackage,
        new_nodes: set,
    ) -> None:
        """
        This adds the modifications to the technosphere matrix for each time-dependent exchange as datapackage elements to a given `bwp.Datapackage`.
        Modifications include:
        1) Exploded processes: new matrix elements for time-explicit consumer and time-explicit producer, representing the temporal edge between them.
        2) Temporal markets: new matrix entries for "temporal markets" and links to the producers in temporally matching background databases. Processes in the background databases are matched on name, reference product and location.
        3) Diagonal entries: ones on the diagonal for new nodes.

        This function also collects the ids of new nodes, temporalized nodes and temporal markets.

        Parameters
        ----------
        row : pd.core.frame
            A row of the timeline DataFrame representing an temporalized edge
        datapackage : bwp.Datapackage
            Append to this datapackage, if available. Otherwise create a new datapackage.
        new_nodes : set
            Set of tuples (node_id, production_amount) to which new node ids are added.

        Returns
        -------
        None but adds elements for this edge to the bwp.Datapackage and stores the ids of new nodes, temporalized nodes and temporal markets.
        """

        if row.consumer == -1:  # functional unit
            new_producer_id = row.time_mapped_producer
            fu_production_amount = bd.get_node(id=row.producer).rp_exchange().amount
            new_nodes.add((new_producer_id, fu_production_amount))
            self.temporalized_process_ids.add(
                new_producer_id
            )  # comes from foreground, so it is a temporalized process
            return

        new_consumer_id = row.time_mapped_consumer
        new_producer_id = row.time_mapped_producer

        previous_producer_id = row.producer
        previous_producer_node = bd.get_node(
            id=previous_producer_id
        )  # in future versions, insead of getting node, just provide list of producer ids

        # Add entry between exploded consumer and exploded producer (not in background database)
        datapackage.add_persistent_vector(
            matrix="technosphere_matrix",
            name=uuid.uuid4().hex,
            data_array=np.array([row.amount], dtype=float),
            indices_array=np.array(
                [(new_producer_id, new_consumer_id)],
                dtype=bwp.INDICES_DTYPE,
            ),
            flip_array=np.array([True], dtype=bool),
        )

        # Check if previous producer comes from background database -> temporal market
        if (
            previous_producer_node["database"]
            in self.database_date_dict_static_only.keys()
        ):
            # Create new edges based on interpolation_weights from the row
            for database, db_share in row.interpolation_weights.items():
                # Get the producer activity in the corresponding background database
                try:
                    producer_id_in_background_db = bd.get_node(
                        **{
                            "database": database,
                            "name": previous_producer_node["name"],
                            "product": previous_producer_node["reference product"],
                            "location": previous_producer_node["location"],
                        }
                    ).id
                except:
                    print(
                        f"Could not find producer in database {database} with name {previous_producer_node['name']}, product {previous_producer_node['reference product']}, location {previous_producer_node['location']}"
                    )
                    raise SystemExit

                # Add entry between exploded producer and producer in background database ("Temporal Market")
                datapackage.add_persistent_vector(
                    matrix="technosphere_matrix",
                    name=uuid.uuid4().hex,
                    data_array=np.array(
                        [db_share], dtype=float
                    ),  # temporal markets produce 1, so shares divide amount between dbs
                    indices_array=np.array(
                        [(producer_id_in_background_db, new_producer_id)],
                        dtype=bwp.INDICES_DTYPE,
                    ),
                    flip_array=np.array([True], dtype=bool),
                )
                self.temporal_market_ids.add(new_producer_id)
                producer_production_amount = (
                    1  # Shares sum up to 1, so production amount is 1
                )

        else:  # comes from foreground, so it is a temporalized process
            self.temporalized_process_ids.add(new_producer_id)

            # Get the production amount of the previous producer if its not a temporal market - needed diagonal matrix entry
            if isinstance(
                previous_producer_node, bd.backends.iotable.proxies.IOTableActivity
            ):
                if len(previous_producer_node.production()) == 1:
                    producer_production_amount = list(
                        previous_producer_node.production()
                    )[0].amount
                else:
                    raise ValueError(
                        "The producer activity is of type IOTableActivity, but has more than one production exchange. This is currently not supported."
                    )
            elif isinstance(previous_producer_node, bd.backends.proxies.Activity):
                producer_production_amount = (
                    bd.get_node(id=row.producer).rp_exchange().amount
                )
            else:
                raise ValueError(
                    f"Can't determine the production amount of the producer activity {previous_producer_node['name']} , as it's of an unknown type."
                )

        # Add newly created producing process to new_nodes
        new_nodes.add((new_producer_id, producer_production_amount))



from datetime import datetime

import bw2data as bd
import pytest

from bw_timex import TimexLCA


# make sure the test db is loaded
def test_nonunitary_db_fixture(nonunitary_db):
    assert len(bd.databases) == 3


@pytest.mark.usefixtures("nonunitary_db")
class TestClass_EV:

    @pytest.fixture(autouse=True)
    def setup_method(self):
        self.node_a = bd.get_node(database="foreground", code="A")

        database_date_dict = {
            "db_2020": datetime.strptime("2020", "%Y"),
            "foreground": "dynamic",
        }

        self.tlca = TimexLCA(
            demand={self.node_a: 1},
            method=("GWP", "example"),
            database_date_dict=database_date_dict,
        )

        self.tlca.build_timeline(
            starting_datetime=datetime.strptime("2024-01-02", "%Y-%m-%d"),
        )
        self.tlca.lci()
        self.tlca.static_lcia()

    def test_timex_lca_score(self):

        expected_score = (
            1 * 1.5 / 3 * 0.5 + 1 * 4 / 7 * 0.9  # A -> Nonunitary  C in bd_2020
        )  # A -> B -> Nonunitary B in foreground

        false_score = (
            1 * 1.5 / 3 * 0.5 + 1 * 4 * 0.9  # A -> Nonunitary C in bd_2020
        )  # A -> B -> Nonunitary B in foreground (not scaled by production amount!)

        print(false_score)

        assert self.tlca.score == expected_score
